from __future__ import annotations

import typing

from flood import spec
import flood


def generate_test_eth_get_block_by_number(
    *,
    rates: typing.Sequence[int],
    duration: int | None = None,
    durations: typing.Sequence[int] | None = None,
    network: str,
    vegeta_kwargs: typing.Mapping[str, str | None] | None = None,
    random_seed: spec.RandomSeed | None = None,
) -> flood.LoadTest:
    n_calls = flood.estimate_call_count(
        rates=rates, duration=duration, durations=durations
    )
    calls = flood.generate_calls_eth_get_block_by_number(
        n_calls=n_calls,
        network=network,
        random_seed=random_seed,
    )
    return flood.create_load_test(
        calls=calls,
        rates=rates,
        duration=duration,
        durations=durations,
    )


# def generate_test_eth_get_block_by_hash(
#     *,
#     rates: typing.Sequence[int],
#     duration: int | None = None,
#     durations: typing.Sequence[int] | None = None,
#     network: str,
#     vegeta_kwargs: typing.Mapping[str, str | None] | None = None,
#     random_seed: spec.RandomSeed | None = None,
# ) -> flood.LoadTest:
#     n_calls = flood.estimate_call_count(
#         rates=rates, duration=duration, durations=durations
#     )
#     calls = flood.generate_calls_eth_get_block_by_hash(
#         n_calls=n_calls,
#         network=network,
#         random_seed=random_seed,
#     )
#     return flood.create_load_test(
#         calls=calls,
#         rates=rates,
#         duration=duration,
#         durations=durations,
#     )



from __future__ import annotations

import typing

import flood


def generate_transaction_hashes(
    n: int,
    network: str,
    random_seed: flood.RandomSeed | None = None,
) -> typing.Sequence[str]:
    return flood.load_samples(
        network=network,
        datatype='transactions',
        n=n,
        random_seed=random_seed,
    )



from __future__ import annotations

import typing

from . import raw_data_spec

if typing.TYPE_CHECKING:
    import polars as pl


def create_samples_dataset(
    *,
    output_dir: str,
    network: str,
    version: str | None = None,
    sizes: typing.Mapping[str, int] | typing.Sequence[str] | None = None,
    datatypes: typing.Sequence[str] | None = None,
    verbose: bool = True,
) -> None:
    # parse inputs
    if version is None:
        version = 'v1_0_0'
    if isinstance(sizes, list):
        sizes = {
            k: v for k, v in raw_data_spec.default_sizes.items() if k in sizes
        }
    elif isinstance(sizes, dict):
        pass
    elif sizes is None:
        sizes = raw_data_spec.default_sizes
    else:
        raise Exception('invalid sizes format: ' + str(type(sizes)))
    if datatypes is None:
        datatypes = raw_data_spec.default_datatypes
    max_size = max(sizes.values())

    # print summary
    if verbose:
        print('creating dataset of raw EVM samples')
        print('- version:', version)
        print('- datatypes:', datatypes)
        print('- network:', network)
        print('- sizes:', sizes)
        if output_dir is not None:
            print('- saving to:', output_dir)
        print()
        print('collecting samples...')

    # collect samples
    samples = _create_raw_samples(
        n=max_size,
        network='ethereum',
        datatypes=datatypes,
    )

    # write sample files
    if verbose:
        print('writing samples to disk...')
    _write_raw_samples(
        samples=samples,
        output_dir=output_dir,
        sizes=sizes,
        network=network,
        version=version,
    )

    if verbose:
        print('sample dataset completed')


def _create_raw_samples(
    n: int,
    network: str,
    datatypes: typing.Sequence[str] | None = None,
) -> typing.Mapping[str, pl.DataFrame]:
    import glob
    import pdp
    from pdp.datasets import contracts
    import polars as pl

    if datatypes is None:
        datatypes = raw_data_spec.default_datatypes

    samples = {}

    # contracts
    if 'contracts' in datatypes:
        contracts_df = contracts.query_contracts(
            columns=['contract_address', 'block_number', 'deployer'],
            network=network,
        ).sort('block_number')
        samples['contracts'] = contracts_df[
            ['contract_address', 'block_number']
        ].sample(n=n)

    # slots
    if 'slots' in datatypes:
        slot_glob = pdp.get_dataset_glob(
            network=network,
            datatype='slots',
        )
        pieces = []
        for file in glob.glob(slot_glob):
            piece = (
                pl.scan_parquet(file)
                .select('contract_address', 'slot', 'first_updated_block')
                .collect()
                .sample(int(n / 10))
            )
            pieces.append(piece)
        samples['slots'] = (
            pl.concat(pieces)
            .rename({'first_updated_block': 'block_number'})
            .sample(n)
        )

    # transactions
    if 'transactions' in datatypes:
        native_transfers_glob = pdp.get_dataset_glob(
            network=network,
            datatype='native_transfers',
        )
        pieces = []
        for file in glob.glob(native_transfers_glob):
            piece = (
                pl.scan_parquet(file)
                .select('transaction_hash', 'block_number')
                .collect()
            )
            if len(piece) > n / 50:
                piece = piece.sample(int(n / 50))
            pieces.append(piece)
        samples['transactions'] = pl.concat(pieces).sample(n)

    # eoas
    if 'eoas' in datatypes:
        native_transfers_glob = pdp.get_dataset_glob(
            network=network,
            datatype='native_transfers',
        )
        pieces = []
        for file in glob.glob(native_transfers_glob):
            piece = (
                pl.scan_parquet(file)
                .select('from_address', 'block_number')
                .unique('from_address', keep='first')
                .collect()
            )
            if len(piece) > n / 50:
                piece = piece.sample(int(n / 50))
            pieces.append(piece)
        eoas_df = (
            pl.concat(pieces)
            .filter(
                ~pl.col('from_address').is_in(contracts_df['contract_address'])
            )
            .unique('from_address')
            .rename({'from_address': 'eoa'})
        )
        samples['eoas'] = eoas_df.sample(n)

    return samples


def _write_raw_samples(
    samples: typing.Mapping[str, pl.DataFrame],
    *,
    output_dir: str,
    network: str,
    version: str,
    sizes: typing.Mapping[str, int] | None = None,
) -> None:
    import os

    path_template = os.path.join(
        output_dir, raw_data_spec.raw_data_file_template
    )
    if sizes is None:
        sizes = raw_data_spec.default_sizes

    for sample_name, sample_data in samples.items():
        for size_name, size in sizes.items():
            # get output path
            output_path = path_template.format(
                network=network,
                datatype=sample_name,
                size=size_name,
                version=version,
            )
            os.makedirs(os.path.dirname(output_path), exist_ok=True)

            # shrink sample to size
            if len(sample_data) == size:
                sized_data = sample_data
            elif len(sample_data) > size:
                sized_data = sample_data[:size]
            else:
                raise Exception('desired size is too big for sample data')

            # write file
            sized_data.write_parquet(output_path)



from __future__ import annotations

import typing

import flood


def plot_load_test_results(
    outputs: typing.Mapping[str, flood.LoadTestOutput],
    test_name: str,
    output_dir: str | None = None,
    latency_yscale_log: bool = True,
    colors: typing.Mapping[str, str] | None = None,
) -> None:
    import os
    import matplotlib.pyplot as plt  # type: ignore
    import toolplot

    toolplot.setup_plot_formatting()

    if output_dir is not None:
        os.makedirs(output_dir, exist_ok=True)

    plt.figure()
    plot_load_test_success(outputs, test_name=test_name, colors=colors)
    if output_dir is not None:
        path = os.path.join(output_dir, 'success.png')
        plt.savefig(path)
    else:
        plt.show()

    plt.figure()
    plot_load_test_throughput(outputs, test_name=test_name, colors=colors)
    if output_dir is not None:
        path = os.path.join(output_dir, 'throughput.png')
        plt.savefig(path)
    else:
        plt.show()

    plt.figure()
    plot_load_test_latencies(
        outputs,
        test_name=test_name,
        yscale_log=latency_yscale_log,
        colors=colors,
    )
    if output_dir is not None:
        path = os.path.join(output_dir, 'latencies.png')
        plt.savefig(path)
    else:
        plt.show()


def plot_load_test_success(
    results: typing.Mapping[str, flood.LoadTestOutput],
    colors: typing.Mapping[str, str] | None = None,
    test_name: str | None = None,
) -> None:
    import matplotlib.pyplot as plt

    plot_load_test_result_metrics(
        results=results,
        metrics=['success'],
        colors=colors,
        test_name=test_name,
        title='Success Rate vs Request Rate\n(higher is better)',
        ylabel='success rate',
        ylim=[-0.03, 1.03],
    )
    plt.legend(loc='center right')


def plot_load_test_throughput(
    results: typing.Mapping[str, flood.LoadTestOutput],
    colors: typing.Mapping[str, str] | None = None,
    test_name: str | None = None,
) -> None:
    import matplotlib.pyplot as plt

    plot_load_test_result_metrics(
        results=results,
        metrics=['throughput'],
        colors=colors,
        test_name=test_name,
        title='Throughput vs Request Rate\n(higher is better)',
        ylabel='throughput\n(responses per second)',
        ymin=0,
    )
    plt.legend(loc='upper left')


def plot_load_test_latencies(
    results: typing.Mapping[str, flood.LoadTestOutput],
    colors: typing.Mapping[
        str,
        str | typing.Sequence[str] | typing.Mapping[str, str],
    ]
    | None = None,
    metrics: typing.Sequence[str] = ['p99', 'p90', 'p50'],
    test_name: str | None = None,
    yscale_log: bool = False,
) -> None:
    import matplotlib.pyplot as plt

    if colors is None:
        colors = dict(zip(results.keys(), flood.plot_colors.values()))

    if yscale_log:
        ymin = None
    else:
        ymin = 0

    plot_load_test_result_metrics(
        results=results,
        metrics=metrics,
        colors=colors,
        test_name=test_name,
        ymin=ymin,
        title='Latency vs Request Rate\n(lower is better)',
        ylabel='latency (seconds)',
        yscale_log=yscale_log,
    )
    plt.legend(loc='upper left')


def plot_load_test_result_metrics(
    results: typing.Mapping[str, flood.LoadTestOutput],
    metrics: typing.Sequence[str],
    *,
    colors: typing.Mapping[
        str,
        str | typing.Sequence[str] | typing.Mapping[str, str],
    ]
    | None = None,
    test_name: str | None = None,
    title: str | None = None,
    ylabel: str | None = None,
    ylim: typing.Sequence[float] | None = None,
    ymin: float | int | None = None,
    yscale_log: bool = False,
) -> None:
    import matplotlib.pyplot as plt
    import toolplot

    if colors is None:
        colors = {
            key: color for key, color in zip(results.keys(), flood.plot_colors)
        }

    for name, result in results.items():
        # determine colors
        result_colors = colors.get(name)
        if isinstance(result_colors, str):
            if result_colors in flood.plot_colors:
                if len(metrics) == 1:
                    result_colors = [flood.plot_colors[result_colors][1]]
                else:
                    result_colors = flood.plot_colors[result_colors]
            else:
                result_colors = [result_colors] * len(metrics)
        elif isinstance(result_colors, list):
            assert len(result_colors) >= len(metrics), 'not enough colors'
        elif isinstance(result_colors, dict):
            for metric in metrics:
                assert metric in result_colors, 'missing color for ' + metric
            result_colors = [result_colors[metric] for metric in metrics]
        else:
            raise Exception('invalid color format')

        # plot
        for zorder, metric, color in zip(
            range(len(metrics)), metrics, result_colors
        ):
            label = name
            if len(metrics) > 1:
                label += ' ' + metric
            plt.plot(
                result['target_rate'],
                result[metric],  # type: ignore
                '.-',
                markersize=20,
                color=color,
                label=label,
                zorder=zorder,
            )

    # set labels
    if yscale_log:
        plt.yscale('log')
    if ylim is not None:
        plt.ylim(*ylim)
    if ymin is not None:
        ylim = plt.ylim()
        plt.ylim([ymin, ylim[1]])  # type: ignore
    xlabel = 'requests per second'
    if test_name is not None:
        xlabel += '\n[' + test_name + ']'
    toolplot.set_labels(
        title=title,
        xlabel=xlabel,
        ylabel=ylabel,
    )
    plt.legend(loc='center right')



from __future__ import annotations

import typing

import flood
from . import raw_data_spec
from . import raw_download_utils

if typing.TYPE_CHECKING:
    import polars as pl


def get_raw_samples_path(
    network: str,
    datatype: str,
    *,
    size: str | None = None,
    version: str = raw_data_spec.raw_data_version,
    samples_dir: str | None = None,
) -> str | None:
    import os

    if samples_dir is None:
        samples_dir = flood.get_flood_samples_dir()

    # get largest_available file present
    if size is None:
        size = 'largest_available'
    if size == 'largest_available':
        import glob

        path_glob = raw_data_spec.raw_data_file_template.format(
            network=network,
            datatype=datatype,
            size='*',
            version=version,
        )
        filepaths = glob.glob(os.path.join(samples_dir, path_glob))
        if len(filepaths) > 0:
            sizes = [(os.stat(path).st_size, path) for path in filepaths]
            return max(sizes)[1]
        else:
            return None

    else:
        filename = raw_data_spec.raw_data_file_template.format(
            network=network,
            datatype=datatype,
            size=size,
            version=version,
        )
        return os.path.join(samples_dir, filename)


def load_raw_samples(
    network: str,
    datatype: str,
    *,
    size: str | None = None,
    version: str = raw_data_spec.raw_data_version,
    samples_dir: str | None = None,
    download_missing: str | None = 'L',
) -> pl.DataFrame:
    path = get_raw_samples_path(
        datatype=datatype,
        network=network,
        size=size,
        version=version,
        samples_dir=samples_dir,
    )
    if path is None:
        if download_missing is not None:
            if size is None:
                size = 'L'
            raw_download_utils.download_raw_data(
                network=network,
                datatypes=[datatype],
                version=version,
                output_dir=samples_dir,
                sizes=[download_missing],
            )
            path = get_raw_samples_path(
                datatype=datatype,
                network=network,
                size=size,
                version=version,
                samples_dir=samples_dir,
            )
            if path is None:
                raise Exception('could not download necessary data')
        else:
            raise Exception('no raw samples found to load')
    return pl.read_parquet(path)


def load_samples(
    network: str,
    datatype: str,
    n: int,
    *,
    size: str | None = None,
    version: str = raw_data_spec.raw_data_version,
    samples_dir: str | None = None,
    binary_convert: bool = True,
    download_missing: bool = True,
    random_seed: flood.RandomSeed | None = None,
) -> typing.Sequence[typing.Any]:
    import polars as pl

    # get path
    path = get_raw_samples_path(
        datatype=datatype,
        network=network,
        size=size,
        version=version,
        samples_dir=samples_dir,
    )
    if path is None:
        # download data if need be
        if download_missing:
            if size is None:
                size = 'L'
            raw_download_utils.download_raw_data(
                network=network,
                datatypes=[datatype],
                version=version,
                output_dir=samples_dir,
                sizes=[size],
            )
            path = get_raw_samples_path(
                datatype=datatype,
                network=network,
                size=size,
                version=version,
                samples_dir=samples_dir,
            )
            if path is None:
                raise Exception('could not download necessary data')
        else:
            raise Exception('no raw samples found to load')

    columns = {
        'contracts': ['contract_address'],
        'eoas': ['eoa'],
        'transactions': ['transaction_hash'],
        'slots': ['contract_address', 'slot'],
    }[datatype]

    df = pl.scan_parquet(path).select(columns).collect()
    if n > len(df):
        import math

        n_copies = math.ceil(n / len(df))
        df = pl.concat(n_copies * [df])
    if n < len(df):
        rng = flood.get_rng(random_seed=random_seed)
        seed = rng.integers(1_000_000_000, size=1)[0]
        df = df.sample(n, shuffle=True, seed=seed)

    for column in df.select(pl.col(pl.Binary)).columns:
        df = df.with_columns(
            ('0x' + pl.col(column).bin.encode('hex')).alias(column)
        )

    if len(columns) == 1:
        return df[columns[0]].to_list()
    else:
        return df.rows()



from __future__ import annotations

import typing

import flood
from flood import spec
from . import address_generators
from . import block_generators
from . import slot_generators
from . import transaction_generators


#
# # blocks
#


def generate_calls_eth_get_block_by_number(
    n_calls: int | None = None,
    *,
    network: str | None = None,
    block_numbers: typing.Sequence[int] | None = None,
    random_seed: spec.RandomSeed | None = None,
) -> typing.Sequence[spec.Call]:
    import ctc.rpc

    if block_numbers is None:
        if n_calls is None:
            raise Exception('must specify more parameters')
        block_numbers = block_generators.generate_block_numbers(
            n=n_calls,
            random_seed=random_seed,
            start_block=0,
            end_block=16_000_000,
            network=network,
        )
    return [
        ctc.rpc.construct_eth_get_block_by_number(block_number=block_number)
        for block_number in block_numbers
    ]


def generate_calls_eth_get_block_by_hash(
    n_calls: int | None = None,
    *,
    network: str | None = None,
    block_hashes: typing.Sequence[str] | None = None,
    random_seed: spec.RandomSeed | None = None,
) -> typing.Sequence[spec.Call]:
    import ctc.rpc

    if block_hashes is None:
        if n_calls is None:
            raise Exception('must specify more parameters')
        block_hashes = block_generators.generate_block_hashes(
            n=n_calls,
            network=network,
            random_seed=random_seed,
        )
    return [
        ctc.rpc.construct_eth_get_block_by_hash(block_hash=block_hash)
        for block_hash in block_hashes
    ]


#
# # addresses
#


def generate_calls_eth_get_eth_balance(
    n_calls: int | None = None,
    *,
    network: str,
    addresses: typing.Sequence[str] | None = None,
    block_numbers: typing.Sequence[int] | None = None,
    random_seed: spec.RandomSeed | None = None,
) -> typing.Sequence[spec.Call]:
    import ctc.rpc

    if block_numbers is None:
        if n_calls is None:
            raise Exception('must specify more parameters')
        block_numbers = block_generators.generate_block_numbers(
            start_block=10_000_000,
            end_block=16_000_000,
            n=n_calls,
            random_seed=random_seed,
            network=network,
        )
    if addresses is None:
        if n_calls is None:
            raise Exception('must specify more parameters')
        addresses = address_generators.generate_contract_addresses(
            n_calls,
            network=network,
            random_seed=random_seed,
        )

    return [
        ctc.rpc.construct_eth_get_balance(
            address=address, block_number=block_number
        )
        for address, block_number in zip(addresses, block_numbers)
    ]


def generate_calls_eth_get_transaction_count(
    n_calls: int | None = None,
    *,
    network: str,
    addresses: typing.Sequence[str] | None = None,
    block_numbers: typing.Sequence[int] | None = None,
    random_seed: spec.RandomSeed | None = None,
) -> typing.Sequence[spec.Call]:
    import ctc.rpc

    if block_numbers is None:
        if n_calls is None:
            raise Exception('must specify more parameters')
        block_numbers = block_generators.generate_block_numbers(
            start_block=10_000_000,
            end_block=16_000_000,
            n=n_calls,
            random_seed=random_seed,
            network=network,
        )
    if addresses is None:
        if n_calls is None:
            raise Exception('must specify more parameters')
        addresses = address_generators.generate_eoas(
            n_calls,
            network=network,
            random_seed=random_seed,
        )

    return [
        ctc.rpc.construct_eth_get_transaction_count(
            from_address=address, block_number=block_number
        )
        for address, block_number in zip(addresses, block_numbers)
    ]


#
# # transactions
#


def generate_calls_eth_get_transaction_by_hash(
    n_calls: int | None = None,
    *,
    network: str,
    transaction_hashes: typing.Sequence[str] | None = None,
    random_seed: spec.RandomSeed | None = None,
) -> typing.Sequence[spec.Call]:
    import ctc.rpc

    if transaction_hashes is None:
        if n_calls is None:
            raise Exception('must specify more parameters')
        transaction_hashes = transaction_generators.generate_transaction_hashes(
            n_calls,
            network=network,
            random_seed=random_seed,
        )
    return [
        ctc.rpc.construct_eth_get_transaction_by_hash(
            transaction_hash=transaction_hash
        )
        for transaction_hash in transaction_hashes
    ]


def generate_calls_eth_get_transaction_receipt(
    n_calls: int | None = None,
    *,
    network: str,
    transaction_hashes: typing.Sequence[str] | None = None,
    random_seed: spec.RandomSeed | None = None,
) -> typing.Sequence[spec.Call]:
    import ctc.rpc

    if transaction_hashes is None:
        if n_calls is None:
            raise Exception('must specify more parameters')
        transaction_hashes = transaction_generators.generate_transaction_hashes(
            n_calls,
            network=network,
            random_seed=random_seed,
        )
    return [
        ctc.rpc.construct_eth_get_transaction_receipt(
            transaction_hash=transaction_hash
        )
        for transaction_hash in transaction_hashes
    ]


#
# # logs
#

_default_contracts = {
    'USDC': '0xa0b86991c6218b36c1d19d4a2e9eb0ce3606eb48',
    'DAI': '0x6b175474e89094c44da98b954eedeac495271d0f',
    'LUSD': '0x5f98805a4e8be255a32880fdec7f6728c6568ba0',
}

_default_event_hashes = {
    'Transfer': '0xddf252ad1be2c89b69c2b068fc378daa952ba7f163c4a11628f55a4df523b3ef',  # noqa: E501
    'Approval': '0x8c5be1e5ebec7d5bd14f71427d1e84f3dd0314c0f7b2291e5b200ac8c7c3b925',  # noqa: E501
}


def generate_calls_eth_get_logs(
    n_calls: int | None = None,
    *,
    contract_address: str | None = None,
    topics: typing.Sequence[str | None] | None = None,
    block_ranges: typing.Sequence[tuple[int, int]] | None = None,
    network: str | None = None,
    random_seed: spec.RandomSeed | None = None,
) -> typing.Sequence[spec.Call]:
    import ctc.rpc

    if contract_address is None:
        contract_address = _default_contracts['USDC']
    if block_ranges is None:
        if n_calls is None:
            raise Exception('must specify more parameters')
        block_ranges = block_generators.generate_block_ranges(
            start_block=10_000_000,
            end_block=16_000_000,
            n=n_calls,
            range_size=100,
            random_seed=random_seed,
            network=network,
        )
    if topics is None:
        topics = [_default_event_hashes['Transfer']]
    return [
        ctc.rpc.construct_eth_get_logs(
            address=contract_address,
            topics=topics,
            start_block=start_block,
            end_block=end_block,
        )
        for start_block, end_block in block_ranges
    ]


#
# # contracts
#


def generate_calls_eth_get_code(
    n_calls: int | None = None,
    *,
    network: str,
    addresses: typing.Sequence[str] | None = None,
    block_numbers: typing.Sequence[int | typing.Literal['latest']]
    | None = None,
    random_seed: spec.RandomSeed | None = None,
) -> typing.Sequence[spec.Call]:
    import ctc.rpc

    if block_numbers is None:
        if n_calls is None:
            raise Exception('must specify more parameters')
        block_numbers = block_generators.generate_block_numbers(
            start_block=10_000_000,
            end_block=16_000_000,
            n=n_calls,
            random_seed=random_seed,
            network=network,
        )
    if addresses is None:
        if n_calls is None:
            raise Exception('must specify more parameters')
        addresses = address_generators.generate_contract_addresses(
            n_calls,
            network=network,
            random_seed=random_seed,
        )
    return [
        ctc.rpc.construct_eth_get_code(
            address=address, block_number=block_number
        )
        for address, block_number in zip(addresses, block_numbers)
    ]


def generate_calls_eth_get_storage_at(
    n_calls: int | None = None,
    *,
    network: str,
    slots: typing.Sequence[tuple[str, str]] | None = None,
    block_numbers: typing.Sequence[int | typing.Literal['latest']]
    | None = None,
    random_seed: spec.RandomSeed | None = None,
) -> typing.Sequence[spec.Call]:
    import ctc.rpc

    if block_numbers is None:
        if n_calls is None:
            raise Exception('must specify more parameters')
        block_numbers = block_generators.generate_block_numbers(
            start_block=10_000_000,
            end_block=16_000_000,
            n=n_calls,
            random_seed=random_seed,
            network=network,
        )
    if slots is None:
        if n_calls is None:
            raise Exception('must specify more parameters')
        slots = slot_generators.generate_slots(
            n_calls, network=network, random_seed=random_seed
        )
    return [
        ctc.rpc.construct_eth_get_storage_at(
            address=address, position=slot, block_number=block_number
        )
        for (address, slot), block_number in zip(slots, block_numbers)
    ]


_default_call_datas = {
    'decimals': '0x313ce567',
    'totalSupply': '0x18160ddd',
    'symbol': '0x95d89b41',
}


def generate_calls_eth_call(
    n_calls: int,
    network: str,
    random_seed: spec.RandomSeed | None = None,
) -> typing.Sequence[spec.Call]:
    import ctc.rpc

    if network != 'ethereum':
        raise Exception('only ethereum supported for eth_call')

    rng = flood.get_rng(random_seed=random_seed)
    contract_addresses = rng.choice(
        list(_default_contracts.values()),
        size=n_calls,
    )
    call_datas = rng.choice(
        list(_default_call_datas.values()),
        size=n_calls,
    )
    block_numbers = block_generators.generate_block_numbers(
        start_block=10_000_000,
        end_block=16_000_000,
        n=n_calls,
        random_seed=random_seed,
        network=network,
    )

    return [
        ctc.rpc.construct_eth_call(
            to_address=contract_address,
            call_data=call_data,
            block_number=block_number,
        )
        for contract_address, call_data, block_number in zip(
            contract_addresses, call_datas, block_numbers
        )
    ]


#
# # traces
#


def generate_calls_trace_block(
    n_calls: int | None = None,
    *,
    block_numbers: typing.Sequence[int] | None = None,
    network: str | None = None,
    random_seed: spec.RandomSeed | None = None,
) -> typing.Sequence[spec.Call]:
    import ctc.rpc

    if block_numbers is None:
        if n_calls is None:
            raise Exception('must specify more parameters')
        block_numbers = block_generators.generate_block_numbers(
            n=n_calls,
            random_seed=0,
            start_block=0,
            end_block=16_000_000,
            network=network,
        )
    return [
        ctc.rpc.construct_trace_block(block_number=block_number)
        for block_number in block_numbers
    ]


def generate_calls_trace_transaction(
    n_calls: int | None = None,
    *,
    transaction_hashes: typing.Sequence[str] | None = None,
    network: str | None = None,
    random_seed: spec.RandomSeed | None = None,
) -> typing.Sequence[spec.Call]:
    import ctc.rpc

    if transaction_hashes is None:
        if n_calls is None:
            raise Exception('must specify more parameters')
        if network is None:
            raise Exception('must specify network')
        transaction_hashes = transaction_generators.generate_transaction_hashes(
            n=n_calls,
            network=network,
            random_seed=random_seed,
        )
    return [
        ctc.rpc.construct_trace_transaction(transaction_hash=transaction_hash)
        for transaction_hash in transaction_hashes
    ]


def generate_calls_trace_replay_block_transactions(
    n_calls: int | None = None,
    *,
    block_numbers: typing.Sequence[int] | None = None,
    network: str | None = None,
    random_seed: spec.RandomSeed | None = None,
) -> typing.Sequence[spec.Call]:
    import ctc.rpc

    if block_numbers is None:
        if n_calls is None:
            raise Exception('must specify more parameters')
        block_numbers = block_generators.generate_block_numbers(
            n=n_calls,
            random_seed=random_seed,
            start_block=0,
            end_block=16_000_000,
            network=network,
        )
    return [
        ctc.rpc.construct_trace_replay_block_transactions(
            block_number=block_number,
            trace_type=['trace'],
        )
        for block_number in block_numbers
    ]


def generate_calls_trace_replay_block_transactions_state_diff(
    n_calls: int | None = None,
    *,
    block_numbers: typing.Sequence[int] | None = None,
    network: str | None = None,
    random_seed: spec.RandomSeed | None = None,
) -> typing.Sequence[spec.Call]:
    import ctc.rpc

    if block_numbers is None:
        if n_calls is None:
            raise Exception('must specify more parameters')
        block_numbers = block_generators.generate_block_numbers(
            n=n_calls,
            random_seed=random_seed,
            start_block=0,
            end_block=16_000_000,
            network=network,
        )
    return [
        ctc.rpc.construct_trace_replay_block_transactions(
            block_number=block_number,
            trace_type=['stateDiff'],
        )
        for block_number in block_numbers
    ]


def generate_calls_trace_replay_block_transactions_vm_trace(
    n_calls: int | None = None,
    *,
    block_numbers: typing.Sequence[int] | None = None,
    network: str | None = None,
    random_seed: spec.RandomSeed | None = None,
) -> typing.Sequence[spec.Call]:
    import ctc.rpc

    if block_numbers is None:
        if n_calls is None:
            raise Exception('must specify more parameters')
        block_numbers = block_generators.generate_block_numbers(
            n=n_calls,
            random_seed=random_seed,
            start_block=0,
            end_block=16_000_000,
            network=network,
        )
    return [
        ctc.rpc.construct_trace_replay_block_transactions(
            block_number=block_number,
            trace_type=['vmTrace'],
        )
        for block_number in block_numbers
    ]


def generate_calls_trace_replay_transaction(
    n_calls: int | None = None,
    *,
    transaction_hashes: typing.Sequence[str] | None = None,
    network: str | None = None,
    random_seed: spec.RandomSeed | None = None,
) -> typing.Sequence[spec.Call]:
    import ctc.rpc

    if transaction_hashes is None:
        if n_calls is None:
            raise Exception('must specify more parameters')
        if network is None:
            raise Exception('must specify network')
        transaction_hashes = transaction_generators.generate_transaction_hashes(
            n=n_calls,
            random_seed=random_seed,
            network=network,
        )
    return [
        ctc.rpc.construct_trace_replay_transaction(
            transaction_hash=transaction_hash,
            trace_type=['trace'],
        )
        for transaction_hash in transaction_hashes
    ]


def generate_calls_trace_replay_transaction_state_diff(
    n_calls: int | None = None,
    *,
    transaction_hashes: typing.Sequence[str] | None = None,
    network: str | None = None,
    random_seed: spec.RandomSeed | None = None,
) -> typing.Sequence[spec.Call]:
    import ctc.rpc

    if transaction_hashes is None:
        if n_calls is None:
            raise Exception('must specify more parameters')
        if network is None:
            raise Exception('must specify network')
        transaction_hashes = transaction_generators.generate_transaction_hashes(
            n=n_calls,
            random_seed=random_seed,
            network=network,
        )
    return [
        ctc.rpc.construct_trace_replay_transaction(
            transaction_hash=transaction_hash,
            trace_type=['stateDiff'],
        )
        for transaction_hash in transaction_hashes
    ]


def generate_calls_trace_replay_transaction_vm_trace(
    n_calls: int | None = None,
    *,
    transaction_hashes: typing.Sequence[str] | None = None,
    network: str | None = None,
    random_seed: spec.RandomSeed | None = None,
) -> typing.Sequence[spec.Call]:
    import ctc.rpc

    if transaction_hashes is None:
        if n_calls is None:
            raise Exception('must specify more parameters')
        if network is None:
            raise Exception('must specify network')
        transaction_hashes = transaction_generators.generate_transaction_hashes(
            n=n_calls,
            random_seed=random_seed,
            network=network,
        )
    return [
        ctc.rpc.construct_trace_replay_transaction(
            transaction_hash=transaction_hash,
            trace_type=['vmTrace'],
        )
        for transaction_hash in transaction_hashes
    ]



from __future__ import annotations

import typing

import flood
from flood import user_io
from flood import spec
from . import load_test_construction
from . import vegeta

if typing.TYPE_CHECKING:
    import multiprocessing


def run_load_tests(
    *,
    node: spec.NodeShorthand | None = None,
    nodes: spec.NodesShorthand | None = None,
    test: spec.LoadTest | None = None,
    tests: typing.Mapping[str, spec.LoadTest] | None = None,
    verbose: bool | int = False,
    include_raw_output: bool = False,
) -> typing.Mapping[str, spec.LoadTestOutput]:
    """run multiple load tests"""
    # parse user_io
    if (node is None) == (nodes is None):
        raise Exception('must specify either node or nodes')
    if (test is None) == (tests is None):
        raise Exception('must specify either test or tests')
    if node is not None:
        node = user_io.parse_node(node)
    if nodes is not None:
        nodes = user_io.parse_nodes(nodes, request_metadata=False)
    tqdm = user_io.outputs._get_tqdm()
    pbar = {
        'leave': False,
        'desc': 'nodes',
        'position': 0,
        'colour': flood.styles['content'],
        # 'disable': not verbose,
        'disable': True,
    }

    results = {}

    # case: single node and single test
    if node is not None and test is not None:
        results[node['name']] = schedule_load_test(
            node=node,
            test=test,
            include_raw_output=include_raw_output,
        )

    # case: single node and multiple tests
    elif node is not None and tests is not None:
        for name, test in tqdm.tqdm(tests.items(), **pbar):
            results[name] = schedule_load_test(
                node=node,
                verbose=verbose,
                test=test,
                include_raw_output=include_raw_output,
            )

    # case: multiple nodes and single tests
    elif nodes is not None and test is not None:
        for name, nd in tqdm.tqdm(nodes.items(), **pbar):
            results[name] = schedule_load_test(
                node=nd,
                verbose=verbose,
                test=test,
                include_raw_output=include_raw_output,
            )

    # case: multiple nodes and multiple tests
    elif nodes is not None and tests is not None:
        for node_name, node in nodes.items():
            for test_name, test in tests.items():
                results[node_name + '__' + test_name] = schedule_load_test(
                    node=node,
                    verbose=verbose,
                    test=test,
                    include_raw_output=include_raw_output,
                )

    # case: invalid input
    else:
        raise Exception('invalid user_io')

    # join any multiprocessing results
    joined = {}
    for name, result in results.items():
        if isinstance(result, dict):
            joined[name] = result
        elif isinstance(result, tuple):
            process, queue = result
            process.join()
            results_path = queue.get()
            with open(results_path, 'r') as f:
                import json

                test_results: spec.SingleRunResultsPayload = json.load(f)
                joined[name] = test_results['results'][name]
        else:
            raise Exception('invalid result type')

    return joined


def schedule_load_test(
    *,
    node: spec.NodeShorthand,
    test: spec.LoadTest | None = None,
    rates: typing.Sequence[int] | None = None,
    calls: typing.Sequence[typing.Any] | None = None,
    duration: int | None = None,
    durations: typing.Sequence[int] | None = None,
    vegeta_kwargs: typing.Mapping[str, str | None] | None = None,
    verbose: bool | int = False,
    include_raw_output: bool = False,
    _pbar_kwargs: typing.Mapping[str, typing.Any] | None = None,
) -> (
    spec.LoadTestOutput
    | str
    | tuple[multiprocessing.Process, multiprocessing.Queue[str]]
):
    """runs local tests synchronously, remote tests asynchronously"""

    node = user_io.parse_node(node)
    if node['remote'] is not None:
        import multiprocessing

        queue: multiprocessing.Queue[str] = multiprocessing.Queue()
        process = multiprocessing.Process(
            target=run_load_test,
            kwargs=dict(
                node=node,
                test=test,
                rates=rates,
                calls=calls,
                duration=duration,
                durations=durations,
                vegeta_kwargs=vegeta_kwargs,
                verbose=verbose,
                include_raw_output=include_raw_output,
                _pbar_kwargs=_pbar_kwargs,
                _container=queue,
            ),
        )
        process.start()
        return (process, queue)
    else:
        return run_load_test(
            node=node,
            test=test,
            rates=rates,
            calls=calls,
            duration=duration,
            durations=durations,
            vegeta_kwargs=vegeta_kwargs,
            verbose=verbose,
            include_raw_output=include_raw_output,
            _pbar_kwargs=_pbar_kwargs,
        )


def run_load_test(
    *,
    node: spec.NodeShorthand,
    test: spec.LoadTest | None = None,
    rates: typing.Sequence[int] | None = None,
    calls: typing.Sequence[typing.Any] | None = None,
    duration: int | None = None,
    durations: typing.Sequence[int] | None = None,
    vegeta_kwargs: typing.Mapping[str, str | None] | None = None,
    verbose: bool | int = False,
    _pbar_kwargs: typing.Mapping[str, typing.Any] | None = None,
    _container: multiprocessing.Queue[str] | None = None,
    include_raw_output: bool = False,
) -> spec.LoadTestOutput | str:
    """run a load test against a single node"""

    # parse user_io
    node = user_io.parse_node(node)
    if test is None:
        if rates is None or calls is None:
            raise Exception('specify rates and calls')
        test = load_test_construction.create_load_test(
            rates=rates,
            calls=calls,
            duration=duration,
            durations=durations,
            vegeta_kwargs=vegeta_kwargs,
        )

    # run tests
    if node.get('remote') is None:
        result: spec.LoadTestOutput | str = _run_load_test_locally(
            node=node,
            test=test,
            verbose=verbose,
            _pbar_kwargs=_pbar_kwargs,
            include_raw_output=include_raw_output,
        )
    else:
        result = _run_load_test_remotely(
            node=node,
            test=test,
            verbose=verbose,
            _pbar_kwargs=_pbar_kwargs,
            include_raw_output=include_raw_output,
        )

    if _container is not None:
        if not isinstance(result, str):
            raise Exception('_container only supportted for remote testing')
        _container.put(result)

    return result


def _run_load_test_locally(
    *,
    node: spec.Node,
    test: spec.LoadTest | None,
    verbose: bool | int = False,
    _pbar_kwargs: typing.Mapping[str, typing.Any] | None = None,
    include_raw_output: bool = False,
) -> spec.LoadTestOutput:
    """run a load test from local node"""
    import datetime
    import toolstr

    if verbose:
        dt = datetime.datetime.now()
        if dt.microsecond >= 500_000:
            dt = dt + datetime.timedelta(
                microseconds=1_000_000 - dt.microsecond
            )
        else:
            dt = dt - datetime.timedelta(microseconds=dt.microsecond)
        timestamp = (
            toolstr.add_style('\[', flood.styles['content'])
            + toolstr.add_style(str(dt), flood.styles['metavar'])
            + toolstr.add_style(']', flood.styles['content'])
        )
        toolstr.print(timestamp + ' Running load test for ' + node['name'])

    # construct progress bar
    if _pbar_kwargs is None:
        _pbar_kwargs = {}
    tqdm = user_io.outputs._get_tqdm()
    tqdm_kwargs = dict(
        leave=False,
        desc='samples',
        position=1,
        colour=flood.styles['content'],
        # disable=not verbose,
        disable=True,
        **_pbar_kwargs,
    )

    # perform tests
    results = []
    for attack in tqdm.tqdm(test, **tqdm_kwargs):
        if verbose:
            dt = datetime.datetime.now()
            if dt.microsecond >= 500_000:
                dt = dt + datetime.timedelta(
                    microseconds=1_000_000 - dt.microsecond
                )
            else:
                dt = dt - datetime.timedelta(microseconds=dt.microsecond)
            timestamp = (
                toolstr.add_style('\[', flood.styles['content'])
                + toolstr.add_style(str(dt), flood.styles['metavar'])
                + toolstr.add_style(']', flood.styles['content'])
            )
            toolstr.print(
                timestamp
                + ' Running attack at rate = '
                + str(attack['rate'])
                + ' rps'
            )

        result = vegeta.run_vegeta_attack(
            url=node['url'],
            calls=attack['calls'],
            duration=attack['duration'],
            rate=attack['rate'],
            vegeta_kwargs=attack['vegeta_kwargs'],
            verbose=verbose >= 2,
            include_raw_output=include_raw_output,
        )
        results.append(result)
        if verbose >= 2:
            print()

    # format output
    keys = results[0].keys()
    return {key: [result[key] for result in results] for key in keys}  # type: ignore # noqa: E501


def _run_load_test_remotely(
    *,
    node: spec.Node,
    test: spec.LoadTest,
    verbose: bool | int = False,
    _pbar_kwargs: typing.Mapping[str, typing.Any] | None = None,
    include_raw_output: bool = False,
) -> str:
    """run a load test from local node"""

    import os
    import subprocess
    import uuid
    import toolstr

    # parse node specification
    remote = node['remote']
    if remote is None:
        raise Exception('not a remote node')

    # save call data, saving methods to preserve ordering in json
    job_id = str(uuid.uuid4())
    tempdir = '/tmp/flood__' + job_id
    os.makedirs(tempdir)
    flood.runners.single_runner.single_runner_io._save_single_run_test(
        test_name='',
        test=test,
        output_dir=tempdir,
    )

    # send call data to remote server
    if verbose:
        import datetime

        dt = datetime.datetime.now()
        if dt.microsecond >= 500_000:
            dt = dt + datetime.timedelta(
                microseconds=1_000_000 - dt.microsecond
            )
        else:
            dt = dt - datetime.timedelta(microseconds=dt.microsecond)
        timestamp = (
            toolstr.add_style('\[', flood.styles['content'])
            + toolstr.add_style(str(dt), flood.styles['metavar'])
            + toolstr.add_style(']', flood.styles['content'])
        )
        node_name = (
            toolstr.add_style('\[', flood.styles['content'])
            + toolstr.add_style('node', flood.styles['metavar'])
            + toolstr.add_style('=', flood.styles['content'])
            + node['name']
            + toolstr.add_style(']', flood.styles['content'])
        )
        toolstr.print(
            timestamp + ' ' + node_name + ' Sending tests to remote node'
        )
    cmd = 'rsync -r ' + tempdir + ' ' + remote + ':' + os.path.dirname(tempdir)
    subprocess.call(cmd.split(' '), stderr=subprocess.DEVNULL)

    # initiate benchmarks
    if verbose:
        dt = datetime.datetime.now()
        if dt.microsecond >= 500_000:
            dt = dt + datetime.timedelta(
                microseconds=1_000_000 - dt.microsecond
            )
        else:
            dt = dt - datetime.timedelta(microseconds=dt.microsecond)
        timestamp = (
            toolstr.add_style('\[', flood.styles['content'])
            + toolstr.add_style(str(dt), flood.styles['metavar'])
            + toolstr.add_style(']', flood.styles['content'])
        )
        toolstr.print(
            timestamp + ' ' + node_name + ' Executing test on remote node'
        )
    cmd_template = "ssh {host} bash -c 'source ~/.profile; python3 -m flood {test} {name}={url} --output {output} --no-figures {extra_kwargs}'"  # noqa: E501
    if include_raw_output:
        extra_kwargs = '--save-raw-output'
    else:
        extra_kwargs = ''
    cmd = cmd_template.format(
        host=remote,
        name=node['name'],
        url=node['url'],
        test=tempdir,
        output=tempdir,
        extra_kwargs=extra_kwargs,
    )
    cmd = cmd.strip()
    subprocess.check_output(cmd.split(' '), stderr=subprocess.DEVNULL)

    # retrieve benchmark results
    if verbose:
        dt = datetime.datetime.now()
        if dt.microsecond >= 500_000:
            dt = dt + datetime.timedelta(
                microseconds=1_000_000 - dt.microsecond
            )
        else:
            dt = dt - datetime.timedelta(microseconds=dt.microsecond)
        timestamp = (
            toolstr.add_style('\[', flood.styles['content'])
            + toolstr.add_style(str(dt), flood.styles['metavar'])
            + toolstr.add_style(']', flood.styles['content'])
        )
        toolstr.print(timestamp + ' ' + node_name + ' Retrieving results')
    results_path = os.path.join(tempdir, 'results.json')
    cmd = 'rsync ' + remote + ':' + results_path + ' ' + results_path
    subprocess.call(cmd.split(' '), stderr=subprocess.DEVNULL)

    return results_path




from __future__ import annotations

import typing

import flood
from flood.runners.single_runner import single_runner_execution
# from flood.runners.multi_runner import multi_runner_execution


def run(
    test_name: str,
    *,
    nodes: flood.NodesShorthand | None,
    random_seed: flood.RandomSeed | None = None,
    verbose: bool | int = True,
    rates: typing.Sequence[int] | None = None,
    duration: int | None = None,
    durations: typing.Sequence[int] | None = None,
    mode: flood.LoadTestMode | None = None,
    vegeta_kwargs: flood.VegetaKwargsShorthand | None = None,
    dry: bool,
    output_dir: str | bool | None = None,
    figures: bool = True,
    metrics: typing.Sequence[str] | None = None,
    include_raw_output: bool = False,
    deep_check: bool = False,
) -> None:
    """generate and run tests against nodes"""
    import os

    # get output_dir
    output_dir = _get_output_dir(output_dir)

    # run test from path
    if os.path.exists(test_name) or '/' in test_name:
        (
            test_name,
            path_spec,
            test,
            nodes,
        ) = _load_old_test_data(test_name=test_name, nodes=nodes)
        return single_runner_execution._run_single(
            rerun_of=path_spec,
            test=test,
            #
            test_name=test_name,
            nodes=nodes,
            random_seed=random_seed,
            dry=dry,
            output_dir=output_dir,
            verbose=verbose,
            metrics=metrics,
            figures=figures,
            include_raw_output=include_raw_output,
            deep_check=deep_check,
        )

    if nodes is None:
        raise Exception('must specify nodes')

    if test_name in flood.get_single_test_generators():
        single_runner_execution._run_single(
            rates=rates,
            duration=duration,
            durations=durations,
            vegeta_kwargs=vegeta_kwargs,
            #
            test_name=test_name,
            nodes=nodes,
            random_seed=random_seed,
            dry=dry,
            output_dir=output_dir,
            verbose=verbose,
            metrics=metrics,
            figures=figures,
            include_raw_output=include_raw_output,
            deep_check=deep_check,
        )
    elif test_name in flood.get_multi_test_generators():
        raise NotImplementedError()
    else:
        raise Exception('invalid test name')


def _get_output_dir(output_dir: str | bool | None) -> str | None:
    import os

    if isinstance(output_dir, bool):
        if output_dir:
            import tempfile

            output_dir = tempfile.mkdtemp()
        else:
            output_dir = None
    if output_dir is not None:
        output_dir = os.path.abspath(os.path.expanduser(output_dir))

    return output_dir


def _load_old_test_data(
    test_name: str, nodes: flood.NodesShorthand | None
) -> tuple[str, str, flood.LoadTest, flood.NodesShorthand]:
    path_spec = test_name

    try:
        test_payload = flood.load_single_run_test_payload(path_spec)
        test = test_payload['test']
        test_name = test_payload['name']
    except Exception:
        raise Exception('invalid test path: ' + str(path_spec))

    # use old nodes if none specified
    if nodes is None:
        results_payload = flood.load_single_run_results_payload(path_spec)
        nodes = results_payload['nodes']

    return (test_name, path_spec, test, nodes)




import os

import pytest


def get_env_value(name):
    value = os.environ.get(name)
    if value in [None, '']:
        raise Exception(name + ' not set')
    return value


@pytest.fixture
def flood_test_local_node_1():
    skip_if_env_nodes_unset()
    return get_env_value('FLOOD_TEST_LOCAL_NODE_1')


@pytest.fixture
def flood_test_local_node_2():
    skip_if_env_nodes_unset()
    return get_env_value('FLOOD_TEST_LOCAL_NODE_2')


@pytest.fixture
def flood_test_remote_node_1():
    skip_if_env_nodes_unset()
    return get_env_value('FLOOD_TEST_REMOTE_NODE_1')


@pytest.fixture
def flood_test_remote_node_2():
    skip_if_env_nodes_unset()
    return get_env_value('FLOOD_TEST_REMOTE_NODE_2')


def skip_if_env_nodes_unset():
    for var in [
        'FLOOD_TEST_LOCAL_NODE_1',
        'FLOOD_TEST_LOCAL_NODE_2',
        'FLOOD_TEST_REMOTE_NODE_1',
        'FLOOD_TEST_REMOTE_NODE_2',
    ]:
        if os.getenv(var) is None:
            pytest.skip(reason=var + ' env var not set')




import subprocess


def test_equality_tests(flood_test_local_node_1, flood_test_local_node_2):
    cmd = 'flood all node1={} node2={} --equality'.format(
        flood_test_local_node_1,
        flood_test_local_node_2,
    )
    subprocess.check_call(cmd.split(' '))




from __future__ import annotations

import typing

from . import raw_data_spec


def download_raw_data(
    *,
    network: str,
    sizes: str | typing.Sequence[str],
    datatypes: typing.Sequence[str] | str | None = None,
    version: str | None = None,
    output_dir: str | None = None,
    only_missing: bool = False,
    verbose: bool = True,
) -> None:
    import os
    import pdp

    # process inputs
    if version is None:
        version = 'v1_0_0'
    if output_dir is None:
        output_dir = get_flood_samples_dir()
    if datatypes is None:
        datatypes = raw_data_spec.default_datatypes
    if isinstance(datatypes, str):
        datatypes = [datatypes]
    if isinstance(sizes, str):
        sizes = [sizes]
    elif isinstance(sizes, list):
        pass
    elif isinstance(sizes, dict):
        sizes = list(sizes.keys())
    else:
        raise Exception()
    url_root = raw_data_spec.raw_data_url_template.format(
        network=network, version=version
    )

    # download each datatype
    for datatype in datatypes:
        for size in sizes:
            filename = raw_data_spec.raw_data_file_template.format(
                datatype=datatype,
                network=network,
                size=size,
                version=version,
            )
            output_path = os.path.join(output_dir, filename)
            url = url_root + filename

            if only_missing and os.path.isfile(output_path):
                continue

            print('downloading', filename)
            pdp.download_file(url=url, output_path=output_path)


def get_flood_samples_dir() -> str:
    import os

    flood_samples_dir = os.environ.get('FLOOD_SAMPLES_DIR')
    if flood_samples_dir is None:
        import warnings

        warnings.warn('FLOOD_SAMPLES_DIR not set in env, defaulting to CWD')
        return os.path.join(os.getcwd(), 'flood_samples')
    else:
        return flood_samples_dir



from __future__ import annotations

import toolcli

import flood

cd_dir_help = {
    'samples': 'FLOOD_SAMPLES_PATH env value',
}


def cd_dir_getter(dirname: str) -> str:
    if dirname == 'samples':
        return flood.get_flood_samples_dir()
    else:
        raise Exception('unknown path: ' + str(dirname))


def run_cli(raw_command: str | None = None) -> None:
    command_index: toolcli.CommandIndex = {
        (): 'flood.cli.root_command',
        ('help',): 'toolcli.command_utils.standard_subcommands.help_command',
        ('ls',): 'flood.cli.ls_command',
        ('print',): 'flood.cli.print_command',
        ('report',): 'flood.cli.report_command',
        ('samples', 'collect'): 'flood.cli.samples_collect_command',
        ('samples', 'download'): 'flood.cli.samples_download_command',
        ('samples', 'ls'): 'flood.cli.samples_ls_command',
        (
            'version',
        ): 'toolcli.command_utils.standard_subcommands.version_command',
        ('cd',): 'toolcli.command_utils.standard_subcommands.cd_command',
    }

    config: toolcli.CLIConfig = {
        'base_command': 'flood',
        'description': flood.__doc__,
        'version': flood.__version__,
        'default_command_sequence': (),
        'root_help_arguments': True,
        # 'root_help_subcommands': False,
        'include_debug_arg': True,
        'style_theme': flood.styles,
        'cd_dir_help': cd_dir_help,
        'cd_dir_getter': cd_dir_getter,
    }

    toolcli.run_cli(
        command_index=command_index,
        config=config,
    )



import subprocess

import pytest

import flood


cmd_templates = {
    'local_bare': 'flood {test_name} {local_node_1} {local_node_2} -d 1 -r 1 2 4',  # noqa: E501
    # 'local_alias': 'flood {test_name} node1={local_node_1} node2={local_node_2} -d 1 -r 1 2 4',  # noqa: E501
    'remote_bare': 'flood {test_name} {remote_node_1} {remote_node_2} -d 1 -r 20 40 60',  # noqa: E501
    # 'remote_alias': 'flood {test_name} node1={remote_node_1} node2={remote_node_2} -d 1 -r 1 2 4',  # noqa: E501
}


@pytest.mark.parametrize('test_name', flood.get_single_test_generators().keys())
@pytest.mark.parametrize('cmd_template', cmd_templates.keys())
def test_load_test(
    test_name,
    cmd_template,
    flood_test_local_node_1,
    flood_test_local_node_2,
    flood_test_remote_node_1,
    flood_test_remote_node_2,
):
    cmd = cmd_templates[cmd_template].format(
        test_name=test_name,
        local_node_1=flood_test_local_node_1,
        local_node_2=flood_test_local_node_2,
        remote_node_1=flood_test_remote_node_1,
        remote_node_2=flood_test_remote_node_2,
    )
    subprocess.check_call(cmd.split(' '))




from __future__ import annotations

import typing

import flood
from . import single_runner_io


def _print_single_run_preamble(
    *,
    test_name: str,
    rates: typing.Sequence[int],
    durations: typing.Sequence[int],
    vegeta_kwargs: flood.VegetaKwargsShorthand | None,
    rerun_of: str | None = None,
    output_dir: str | None,
) -> None:
    import os
    import toolstr

    _print_single_run_preamble_copy(
        test_name=test_name,
        rerun_of=rerun_of,
        rates=rates,
        durations=durations,
        vegeta_kwargs=vegeta_kwargs,
        output_dir=output_dir,
    )
    if output_dir is not None:
        summary_path = os.path.join(output_dir, 'summary.txt')
        with toolstr.write_stdout_to_file(summary_path):
            _print_single_run_preamble_copy(
                test_name=test_name,
                rerun_of=rerun_of,
                rates=rates,
                durations=durations,
                vegeta_kwargs=vegeta_kwargs,
                output_dir=output_dir,
            )


def _print_single_run_preamble_copy(
    *,
    test_name: str,
    rates: typing.Sequence[int],
    durations: typing.Sequence[int],
    vegeta_kwargs: flood.VegetaKwargsShorthand | None,
    rerun_of: str | None = None,
    output_dir: str | None,
) -> None:
    import toolstr

    toolstr.print_text_box(
        toolstr.add_style('Load test: ' + test_name, flood.styles['metavar']),
        style=flood.styles['content'],
    )
    toolstr.print_bullet(key='sample rates', value=rates, styles=flood.styles)
    if len(set(durations)) == 1:
        toolstr.print_bullet(
            key='sample duration',
            value=durations[0],
            styles=flood.styles,
        )
    else:
        toolstr.print_bullet(
            key='sample durations', value=durations, styles=flood.styles
        )
    if vegeta_kwargs is None or len(vegeta_kwargs) == 0:
        toolstr.print_bullet(key='extra args', value=None, styles=flood.styles)

    if rerun_of is not None:
        toolstr.print_bullet(
            key='rerun of', value=rerun_of, styles=flood.styles
        )
    toolstr.print_bullet(
        key='output directory', value=output_dir, styles=flood.styles
    )
    print()


def _print_run_start() -> None:
    import datetime
    import toolstr

    print()
    print()
    toolstr.print_header(
        'Running load tests...',
        style=flood.styles['content'],
        text_style=flood.styles['metavar'],
    )
    dt = datetime.datetime.now()
    if dt.microsecond >= 500_000:
        dt = dt + datetime.timedelta(microseconds=1_000_000 - dt.microsecond)
    else:
        dt = dt - datetime.timedelta(microseconds=dt.microsecond)
    timestamp = (
        toolstr.add_style('\[', flood.styles['content'])
        + toolstr.add_style(str(dt), flood.styles['metavar'])
        + toolstr.add_style(']', flood.styles['content'])
    )
    toolstr.print(timestamp + ' Starting')


def _print_single_run_conclusion(
    *,
    output_dir: str | None,
    results: typing.Mapping[str, flood.LoadTestOutput],
    metrics: typing.Sequence[str] | None,
    verbose: bool | int,
    figures: bool,
) -> None:
    _print_single_run_conclusion_copy(
        output_dir=output_dir,
        results=results,
        metrics=metrics,
        verbose=verbose,
        figures=figures,
    )
    if output_dir is not None:
        import os
        import toolstr

        summary_path = os.path.join(output_dir, 'summary.txt')
        with toolstr.write_stdout_to_file(summary_path, mode='a'):
            _print_single_run_conclusion_copy(
                output_dir=output_dir,
                results=results,
                metrics=metrics,
                verbose=verbose,
                figures=figures,
            )


def _print_single_run_conclusion_copy(
    output_dir: str | None,
    results: typing.Mapping[str, flood.LoadTestOutput],
    metrics: typing.Sequence[str] | None,
    verbose: bool | int,
    figures: bool,
) -> None:
    import datetime
    import os
    import toolstr

    dt = datetime.datetime.now()
    if dt.microsecond >= 500_000:
        dt = dt + datetime.timedelta(microseconds=1_000_000 - dt.microsecond)
    else:
        dt = dt - datetime.timedelta(microseconds=dt.microsecond)
    timestamp = (
        toolstr.add_style('\[', flood.styles['content'])
        + toolstr.add_style(str(dt), flood.styles['metavar'])
        + toolstr.add_style(']', flood.styles['content'])
    )
    toolstr.print(timestamp + ' Load tests completed.')

    # print message about metrics file
    if output_dir is not None:
        test_path = single_runner_io.get_single_run_test_path(
            output_dir=output_dir
        )
        result_path = single_runner_io.get_single_run_results_path(
            output_dir=output_dir
        )
        figures_path = single_runner_io.get_single_run_figures_path(
            output_dir=output_dir
        )

        print()
        print()
        flood.print_header('Saving results...')
        toolstr.print_bullet(
            key=os.path.relpath(test_path, output_dir),
            value='',
            colon_str='',
            styles=flood.styles,
        )
        toolstr.print_bullet(
            key=os.path.relpath(result_path, output_dir),
            value='',
            colon_str='',
            styles=flood.styles,
        )
        if figures:
            toolstr.print_bullet(
                key=os.path.relpath(figures_path, output_dir),
                value='',
                colon_str='',
                styles=flood.styles,
            )

    # decide metrics
    if metrics is None:
        metrics = ['success', 'throughput', 'p90']

    # print metrics
    print()
    print()
    flood.print_header('Summarizing performance metrics...')
    if verbose > 1:
        toolstr.print_bullet(
            key='metrics shown below',
            value=', '.join(metrics),
            styles=flood.styles,
        )
        example_result = list(results.values())[0]
        additional = [
            key for key in example_result.keys() if key not in metrics
        ]
        toolstr.print_bullet(
            key='additional metrics available',
            value=', '.join(additional),
            styles=flood.styles,
        )

    # print metric values
    print()
    flood.print_metric_tables(
        results=results,
        metrics=metrics,
        indent=4,
    )




from __future__ import annotations

import typing

import flood


#
# # bookkeeping
#


def get_single_test_generators() -> (
    typing.Mapping[str, flood.LoadTestGenerator]
):
    return {
        get_test_generator_display_name(item): item  # type: ignore
        for item in dir(flood)
        if item.startswith('generate_test_')
    }


def get_multi_test_generators() -> (
    typing.Mapping[str, flood.MultiLoadTestGenerator]
):
    return {
        get_test_generator_display_name(item, multi=True): item  # type: ignore
        for item in dir(flood)
        if item.startswith('generate_tests_')
    }


def get_test_generator(test_name: str) -> flood.LoadTestGenerator:
    function_name = get_test_generator_function_name(test_name)
    if hasattr(flood, function_name):
        return getattr(flood, function_name)  # type: ignore
    else:
        raise Exception()


def get_tests_generator(test_name: str) -> flood.MultiLoadTestGenerator:
    function_name = get_test_generator_function_name(test_name)
    if hasattr(flood, function_name):
        return getattr(flood, function_name)  # type: ignore
    else:
        raise Exception()


def get_test_generator_display_name(
    test: str | flood.LoadTestGenerator,
    multi: bool = False,
) -> str:
    if multi:
        prefix = 'generate_tests_'
    else:
        prefix = 'generate_test_'

    if not isinstance(test, str):
        import types

        if not isinstance(test, types.FunctionType):
            raise Exception('should be str or function')
        test = test.__name__

    if not test.startswith(prefix):
        raise Exception()
    test = test[len(prefix) :]
    head, tail = test.split('_', 1)
    test = head + '_' + _snake_case_to_camel_case(tail)
    return test


def get_test_generator_function_name(
    display_name: str, multi: bool = False
) -> str:
    if multi:
        prefix = 'generate_tests_'
    else:
        prefix = 'generate_test_'

    function_name = prefix + _camel_case_to_snake_case(display_name)
    return function_name


def _camel_case_to_snake_case(string: str) -> str:
    # adapted from https://stackoverflow.com/a/1176023
    import re

    return re.sub(r'(?<!^)(?=[A-Z])', '_', string).lower()


def _snake_case_to_camel_case(string: str) -> str:
    pieces = string.split('_')
    return pieces[0] + ''.join(piece.title() for piece in pieces[1:])


#
# # generation
#


def generate_test(
    *,
    test_name: str,
    random_seed: flood.RandomSeed | None = None,
    rates: typing.Sequence[int] | None = None,
    durations: typing.Sequence[int] | None = None,
    vegeta_kwargs: flood.VegetaKwargsShorthand | None = None,
    network: str,
    output_dir: str | None = None,
) -> flood.LoadTest:
    if test_name is None:
        raise Exception('must specify test_name')
    test_generator = get_test_generator(test_name)
    test = test_generator(
        rates=rates,
        durations=durations,
        vegeta_kwargs=vegeta_kwargs,
        network=network,
        random_seed=random_seed,
    )
    if output_dir is not None:
        flood.runners.single_runner.single_runner_io._save_single_run_test(
            test_name=test_name, output_dir=output_dir, test=test
        )
    return test


def generate_tests(
    *,
    test_name: str,
    random_seed: flood.RandomSeed | None = None,
    rates: typing.Sequence[int] | None = None,
    durations: typing.Sequence[int] | None = None,
    vegeta_kwargs: flood.VegetaKwargsShorthand | None = None,
    network: str,
    output_dir: str | None = None,
    common_parameters: typing.Mapping[str, typing.Any] | None = None,
    specific_parameters: typing.Mapping[str, typing.Mapping[str, typing.Any]]
    | None = None,
) -> typing.Mapping[str, flood.LoadTest]:
    if test_name is None:
        raise Exception('must specify test_name')
    test_generator = get_tests_generator(test_name)
    tests = test_generator(
        rates=rates,
        durations=durations,
        vegeta_kwargs=vegeta_kwargs,
        network=network,
        random_seed=random_seed,
        common_parameters=common_parameters,
        specific_parameters=specific_parameters,
    )
    if output_dir is not None:
        flood.runners.multi_runner.multi_runner_io._save_multi_run_tests(
            output_dir=output_dir,
            tests=tests,
        )
    return tests


# def generate_tests(
#     test_name: str,
#     constants: typing.Mapping[str, typing.Any] | None = None,
#     variables: typing.Mapping[
#         str, typing.Mapping[str, typing.Sequence[typing.Any]]
#     ]
#     | None = None,
# ) -> typing.Mapping[str, flood.LoadTest]:
#     """variables is in format {variable_name: {test_name: variable_value}}"""
#     import itertools

#     if constants is None:
#         constants = {}
#     if variables is None:
#         variables = {}

#     # compute input variables
#     tests_kwargs = []
#     combinations = itertools.product(*variables.items())
#     for combination in combinations:
#         test_kwargs = dict(combination, **constants)
#         tests_kwargs.append(test_kwargs)

#     # use test generator function
#     test_generator = get_test_generator(test_name)
#     tests = []
#     for test_kwargs in tests_kwargs:
#         test = test_generator(**test_kwargs)
#         tests.append(test)

#     return tests




from __future__ import annotations

import typing

import flood


default_stress_test_rates = [1, 4, 16, 64, 256, 512]
default_stress_test_duration = 30
default_spike_rates = {
    'before': 8,
    'during': 2048,
    'after': 8,
}
default_spike_durations = {
    'before': 30,
    'during': 10,
    'after': 30,
}
default_soak_test_rate = 100
default_soak_test_duration = 24 * 60 * 60


def generate_timings(
    rates: typing.Sequence[int] | None = None,
    duration: int | None = None,
    durations: typing.Sequence[int] | None = None,
    mode: flood.LoadTestMode | None = None,
) -> tuple[typing.Sequence[int], typing.Sequence[int]]:
    """create rates and durations for test"""

    if mode is None:
        mode = 'stress'

    if mode == 'stress':
        return _generate_timings_for_stress_test(
            rates=rates,
            duration=duration,
            durations=durations,
        )
    elif mode == 'soak':
        return _generate_timings_for_soak_test(
            rates=rates,
            duration=duration,
            durations=durations,
        )
    elif mode == 'spike':
        return _generate_timings_for_spike_test(
            rates=rates,
            duration=duration,
            durations=durations,
        )
    else:
        raise Exception('unknown mode: ' + str(mode))


def _generate_timings_for_stress_test(
    rates: typing.Sequence[int] | None = None,
    duration: int | None = None,
    durations: typing.Sequence[int] | None = None,
) -> tuple[typing.Sequence[int], typing.Sequence[int]]:
    if rates is None:
        rates = default_stress_test_rates
    if durations is None:
        if duration is None:
            duration = default_stress_test_duration
        durations = [duration] * len(rates)
    return rates, durations


def _generate_timings_for_spike_test(
    rates: typing.Sequence[int] | None = None,
    duration: int | None = None,
    durations: typing.Sequence[int] | None = None,
) -> tuple[typing.Sequence[int], typing.Sequence[int]]:
    if rates is None:
        rates = [
            default_spike_rates['before'],
            default_spike_rates['during'],
            default_spike_rates['after'],
        ]
    else:
        if len(rates) == 1:
            rates = [
                default_spike_rates['before'],
                rates[0],
                default_spike_rates['after'],
            ]
        elif len(rates) == 3:
            pass
        else:
            raise Exception(
                'must specify eith a single spike rate or a [before, during, after] triplet of rates'  # noqa: E501
            )

    if duration is not None:
        durations = [
            default_spike_durations['before'],
            duration,
            default_spike_durations['after'],
        ]
    elif durations is not None:
        if len(durations) == 1:
            durations = [
                default_spike_durations['before'],
                durations[0],
                default_spike_durations['after'],
            ]
        elif len(durations) == 3:
            pass
        else:
            raise Exception(
                'must specify eith a single spike duration or a [before, during, after] triplet of rates'  # noqa: E501
            )
    else:
        durations = [
            default_spike_durations['before'],
            default_spike_durations['during'],
            default_spike_durations['after'],
        ]

    return rates, durations


def _generate_timings_for_soak_test(
    rates: typing.Sequence[int] | None = None,
    duration: int | None = None,
    durations: typing.Sequence[int] | None = None,
) -> tuple[typing.Sequence[int], typing.Sequence[int]]:
    if rates is not None:
        if len(rates) != 1:
            raise Exception('must specify 1 rate for soak test')
    else:
        rates = [default_soak_test_rate]

    if durations is None:
        if duration is not None:
            durations = [duration]
        else:
            durations = [default_soak_test_duration]
    else:
        if len(durations) != 1:
            raise Exception('must specify 1 duration for soak test')

    return rates, durations



from __future__ import annotations

import os
import toolcli
import toolstr

import flood


def get_command_spec() -> toolcli.CommandSpec:
    return {
        'f': samples_ls_command,
        'help': 'list raw EVM data samples that are downloaded and downloadable',  # noqa: E501
        'args': [],
    }


def samples_ls_command() -> None:
    version = 'v1_0_0'

    # parse contents of flood samples dir
    samples_dir = flood.get_flood_samples_dir()
    files_by_network: dict[str, list[str]] = {}
    for filename in os.listdir(samples_dir):
        if filename.endswith('parquet'):
            network = filename.split('_')[0]
            files_by_network.setdefault(network, [])
            files_by_network[network].append(filename)
    for network in list(files_by_network.keys()):
        files_by_network[network] = sorted(files_by_network[network])

    for network in files_by_network.keys():
        toolstr.print_text_box(network + ' sample data')
        print()
        print(samples_dir)
        print()
        toolstr.print_header('Local Files')
        for filename in files_by_network[network]:
            print('-', filename)

        missing = []
        for size in flood.default_sizes:
            for datatype in flood.default_datatypes:
                filename = flood.raw_data_file_template.format(
                    network=network,
                    size=size,
                    datatype=datatype,
                    version=version,
                )
                if filename not in files_by_network[network]:
                    missing.append(filename)

        if len(missing) > 0:
            print()
            toolstr.print_header('Available for download')
            for filename in missing:
                print('-', filename)
            print()
            print('download using `flood samples download` command')



from __future__ import annotations

import typing

import flood

if typing.TYPE_CHECKING:
    import nbformat


def create_load_test_report(
    test_paths: typing.Sequence[str],
    output_dir: str,
    metrics: typing.Sequence[str] | None = None,
) -> None:
    import os

    ipynb_path = os.path.join(output_dir, 'report.ipynb')
    _create_load_test_report_ipynb(
        test_paths=test_paths,
        output_path=ipynb_path,
        metrics=metrics,
    )

    _create_load_test_report_html(ipynb_path=ipynb_path)


def print_load_test_summary(test: flood.LoadTest) -> None:
    import toolstr

    parsed = flood.parse_test_data(test)
    rates = parsed['rates']
    durations = parsed['durations']
    vegeta_kwargs = parsed['vegeta_kwargs']

    toolstr.print_bullet(key='sample rates', value=rates, styles=flood.styles)
    if len(set(durations)) == 1:
        toolstr.print_bullet(
            key='sample duration',
            value=durations[0],
            styles=flood.styles,
        )
    else:
        toolstr.print_bullet(
            key='sample durations', value=durations, styles=flood.styles
        )
    if vegeta_kwargs is None or len(vegeta_kwargs) == 0:
        toolstr.print_bullet(key='extra args', value=None, styles=flood.styles)


def _create_load_test_report_html(
    ipynb_path: str,
) -> None:
    import subprocess

    cmd = 'python3 -m nbconvert --to html --execute ' + ipynb_path
    subprocess.call(cmd.split(' '))


def _create_load_test_report_ipynb(
    test_paths: typing.Sequence[str],
    output_path: str,
    metrics: typing.Sequence[str] | None = None,
) -> None:
    from flood.user_io import notebook_io

    if metrics is None:
        metrics = ['success', 'throughput', 'p50', 'p90', 'p99']

    test_paths_map = {}
    for test_path in test_paths:
        payload = flood.load_single_run_test_payload(test_path)
        name = payload['name']
        test_paths_map[name] = test_path

    inputs = {
        'test_paths': test_paths_map,
        'metrics': metrics,
        'test_names': list(test_paths_map.keys()),
    }

    notebook_io.create_notebook(
        cell_templates=_report_template_cells,
        output_path=output_path,
        inputs=inputs,
    )


_header_template = """# `flood` Load Test Report


{toc}

### Report Generation

*This report was generated by `flood`. See the `flood report --help` command for report generation options. This report can be executed as a Python notebook using the `.ipynb` version of this file.*"""  # noqa: E501


_toc_template = """### Contents
1. [Test Summary](#Test-Summary)
2. [Tests](#Tests)
{toc_lines}"""


def _create_header_cell(test_names: typing.Sequence[str]) -> str:
    toc = _create_test_toc(test_names=test_names)
    return _header_template.format(toc=toc)


def _create_test_toc(test_names: typing.Sequence[str]) -> str:
    line_template = '. [Test: {test_name}](#Test:-{test_name})\n'
    toc = ''
    for t, test_name in enumerate(test_names):
        toc += '    ' + str(t + 1) + line_template.format(test_name=test_name)
    toc = toc.rstrip()
    return _toc_template.format(toc_lines=toc)


def _create_parameters_cell(
    test_paths: typing.Mapping[str, str],
    metrics: typing.Sequence[str],
) -> str:
    parameters = '# parameters\n\n'

    parameters += 'test_paths = ' + (
        repr(test_paths)
        .replace(',', ',\n   ')
        .replace('{', '{\n    ')
        .replace('}', ',\n}')
    )

    parameters += '\n\nmetrics = ' + str(metrics)

    return parameters


def _create_test_chunks(
    test_names: typing.Sequence[str],
) -> typing.Sequence[nbformat.notebooknode.NotebookNode]:
    return [
        cell
        for test_name in test_names
        for cell in _create_test_chunk(test_name)
    ]


def _create_test_chunk(
    test_name: str,
) -> typing.Sequence[nbformat.notebooknode.NotebookNode]:
    from flood.user_io import notebook_io

    return notebook_io.create_cells(
        cell_templates=_test_template_cells,
        inputs={'test_name': test_name},
    )


if typing.TYPE_CHECKING:
    from flood.user_io import notebook_io

_report_template_cells: notebook_io.NotebookTemplate = [
    {
        # header
        'type': 'markdown',
        'f': _create_header_cell,
        'inputs': ['test_names'],
    },
    {
        # imports
        'type': 'code',
        'content': """
            import IPython
            import polars as pl
            import toolstr
            import tooltime

            import flood

            flood.styles = {{}}
        """,
        'inputs': [],
    },
    {
        # parameters
        'type': 'code',
        'f': _create_parameters_cell,
        'inputs': ['test_paths', 'metrics'],
    },
    {
        # load data
        'type': 'code',
        'content': """
            # load data

            test_payloads = {{
                test_name: flood.load_single_run_test_payload(test_path)
                for test_name, test_path in test_paths.items()
            }}

            results_payloads = {{
                test_name: flood.load_single_run_results_payload(output_dir=test_path)
                for test_name, test_path in test_paths.items()
            }}
        """,  # noqa: E501
        'inputs': [],
    },
    {
        # Test summary header
        'type': 'markdown',
        'content': '# Test Summary',
        'inputs': [],
    },
    {
        # test list
        'type': 'code',
        'content': """
            # test list

            toolstr.print_text_box('Tests')
            for t, test_name in enumerate(results_payloads.keys()):
                print(str(t + 1) + '.', test_name)
        """,
        'inputs': [],
    },
    {
        # test durations
        'type': 'code',
        'content': """
            # test durations

            time_per_test = {{}}
            time_per_condition = {{}}

            for test_name in results_payloads:
                results = results_payloads[test_name]["results"]
                time_per_test[test_name] = 0
                for condition_name in results.keys():
                    time_per_condition.setdefault(condition_name, 0)
                    time = sum(results[condition_name]["actual_duration"]) + sum(
                        results[condition_name]["final_wait_time"]
                    )
                    time_per_test[test_name] += time
                    time_per_condition[condition_name] += time

            toolstr.print_text_box('Total time')
            toolstr.print(tooltime.timelength_to_phrase(int(sum(time_per_test.values()))))
            print()

            toolstr.print_text_box('Total time per test')
            rows = list(time_per_test.items())
            toolstr.print_table(rows, labels=['test', 'time (s)'])

            toolstr.print_text_box('Total time per condition')
            rows = list(time_per_condition.items())
            toolstr.print_table(rows, labels=['condition', 'time (s)'])
        """,  # noqa: E501
        'inputs': [],
    },
    {
        # Tests header
        'type': 'markdown',
        'content': '# Tests',
        'inputs': [],
    },
    {
        # Tests chunks
        'type': 'chunk',
        'f': _create_test_chunks,
        'inputs': ['test_names'],
    },
]


_test_template_cells: notebook_io.NotebookTemplate = [
    {
        # test header
        'type': 'markdown',
        'content': '# Test: {test_name}',
        'inputs': ['test_name'],
    },
    {
        # load test results
        'type': 'code',
        'content': """
            # load test results

            test_name = '{test_name}'
            results_payload = results_payloads[test_name]
            results = results_payload['results']
        """,
        'inputs': ['test_name'],
    },
    {
        # show test metadata
        'type': 'code',
        'content': """
            # show test metadata

            toolstr.print_text_box(test_name + ' parameters')
            flood.print_load_test_summary(results_payload['test'])
            toolstr.print('- nodes tested:')
            nodes_df = pl.from_records(list(results_payload['nodes'].values()))
            toolstr.print_dataframe_as_table(nodes_df)
        """,
        'inputs': [],
    },
    {
        # show result tables
        'type': 'code',
        'content': """
            # show result tables

            flood.print_metric_tables(results, metrics=metrics, comparison=True)
        """,
        'inputs': [],
    },
    {
        # show result figures
        'type': 'code',
        'content': """
            # show result figures

            colors = flood.get_nodes_plot_colors(nodes=results_payload['nodes'])
            flood.plot_load_test_results(
                test_name=test_name,
                outputs=results,
                latency_yscale_log=True,
                colors=colors,
            )
        """,
        'inputs': [],
    },
    {
        # show errors
        'type': 'code',
        'content': """
            # show errors

            toolstr.print_text_box('Error messages present in each test')
            unique_errors = {{}}
            for n, name in enumerate(results.keys()):
                unique_errors.setdefault(name, set())
                unique_errors[name] |= {{
                    error for error_list in results[name]['errors'] for error in error_list
                }}
                print(name)
                for error in unique_errors[name]:
                    print('-', error)
                if n != len(results) - 1:
                    print()
        """,  # noqa: E501
        'inputs': [],
    },
    {
        # show complete results
        'type': 'code',
        'content': """
            # show complete results

            for name in results.keys():
                toolstr.print_text_box(name + " Complete Results")
                df = pl.DataFrame(results[name])
                df = df.drop(
                    "status_codes",
                    "errors",
                    "first_request_timestamp",
                    "last_request_timestamp",
                    "last_response_timestamp",
                )
                IPython.display.display(df)
        """,
        'inputs': [],
    },
]



"""generic functions for creating notebooks"""
from __future__ import annotations

import typing

import nbformat


if typing.TYPE_CHECKING:

    class NotebookCellTemplateContent(typing.TypedDict):
        type: typing.Literal['markdown', 'code']
        content: str
        inputs: typing.Sequence[str]

    class NotebookCellTemplateFunction(typing.TypedDict):
        type: typing.Literal['markdown', 'code']
        f: typing.Callable[..., str]
        inputs: typing.Sequence[str]

    NotebookCellTemplate = typing.Union[
        NotebookCellTemplateContent,
        NotebookCellTemplateFunction,
    ]

    class NotebookChunkTemplate(typing.TypedDict):
        type: typing.Literal['chunk']
        f: typing.Callable[
            ..., typing.Sequence[nbformat.notebooknode.NotebookNode]
        ]
        inputs: typing.Sequence[str]

    NotebookCellPrecursor = typing.Union[
        NotebookCellTemplate,
        NotebookChunkTemplate,
        nbformat.notebooknode.NotebookNode,
    ]
    NotebookTemplate = typing.Sequence[NotebookCellPrecursor]


def create_cell(
    cell_template: NotebookCellTemplate,
    inputs: typing.Mapping[str, typing.Any],
) -> nbformat.notebooknode.NotebookNode:
    # create cell content
    if 'content' in cell_template:
        content = cell_template['content']  # type: ignore
        if 'inputs' in cell_template:
            format_inputs = {
                input: inputs[input] for input in cell_template['inputs']
            }
            content = content.format(**format_inputs)
    elif 'f' in cell_template:
        f_inputs = {input: inputs[input] for input in cell_template['inputs']}
        content = cell_template['f'](**f_inputs)
    else:
        raise Exception('invalid cell format')

    # left justify as needed
    lines = content.split('\n')
    if lines[0] == '':
        left_spaces = [
            len(line) - len(line.lstrip(' '))
            for line in lines
            if not set(line).issubset({' '})
        ]
        cut = min(left_spaces[1:])
        content = '\n'.join(line[cut:] for line in lines)

    content = content.strip()

    # create cell object
    cell_obj: nbformat.notebooknode.NotebookNode
    if cell_template['type'] == 'markdown':
        cell_obj = nbformat.v4.new_markdown_cell(source=content)  # type: ignore
    elif cell_template['type'] == 'code':
        cell_obj = nbformat.v4.new_code_cell(source=content)  # type: ignore
    else:
        raise Exception('invalid cell type')

    return cell_obj


def create_cell_chunk(
    chunk_template: NotebookChunkTemplate,
    inputs: typing.Mapping[str, typing.Any],
) -> typing.Sequence[nbformat.notebooknode.NotebookNode]:
    if 'f' in chunk_template:
        chunk_inputs = {
            input: inputs[input] for input in chunk_template['inputs']
        }
        cell_templates = chunk_template['f'](**chunk_inputs)
    else:
        raise Exception('invalid chunk template format')

    return create_cells(cell_templates=cell_templates, inputs=inputs)


def create_cells(
    cell_templates: typing.Sequence[NotebookCellPrecursor],
    inputs: typing.Mapping[str, typing.Any],
) -> typing.Sequence[nbformat.notebooknode.NotebookNode]:
    cell_objs = []
    for cell_template in cell_templates:
        if isinstance(cell_template, nbformat.notebooknode.NotebookNode):
            cell_objs.append(cell_template)
        elif cell_template['type'] == 'chunk':
            cell_objs += create_cell_chunk(
                chunk_template=cell_template,
                inputs=inputs,
            )
        else:
            cell_obj = create_cell(
                cell_template=cell_template,
                inputs=inputs,
            )
            cell_objs.append(cell_obj)
    return cell_objs


def create_notebook(
    cell_templates: NotebookTemplate,
    output_path: str,
    inputs: typing.Mapping[str, typing.Any] | None = None,
) -> None:
    # create cells
    if inputs is None:
        inputs = {}
    cells = create_cells(cell_templates=cell_templates, inputs=inputs)

    # create notebook object
    notebook = nbformat.v4.new_notebook()  # type: ignore
    notebook['cells'] = cells

    # write notebook to file
    with open(output_path, 'w') as f:
        nbformat.write(notebook, f)  # type: ignore



from __future__ import annotations

import typing

import flood
from flood import block_generators


def get_all_equality_tests(
    start_block: int = 10_000_000,
    end_block: int = 16_000_000,
    range_size: int = 100,
    random_seed: flood.RandomSeed | None = None,
) -> typing.Sequence[flood.EqualityTest]:
    return list(
        get_vanilla_equality_tests(
            start_block=start_block,
            end_block=end_block,
            range_size=range_size,
            random_seed=random_seed,
        )
    ) + list(
        get_trace_equality_tests(
            start_block=start_block,
            end_block=end_block,
            random_seed=random_seed,
        )
    )


def get_vanilla_equality_tests(
    start_block: int = 10_000_000,
    end_block: int = 16_000_000,
    range_size: int = 100,
    random_seed: flood.RandomSeed | None = None,
) -> typing.Sequence[flood.EqualityTest]:
    import ctc.rpc

    start_block, end_block = block_generators.generate_block_ranges(
        n=1,
        range_size=range_size,
        start_block=start_block,
        end_block=end_block,
        random_seed=random_seed,
    )[0]

    return [
        (
            'eth_getChainId',
            ctc.rpc.construct_eth_chain_id,
            [],
            {},
        ),
        (
            'eth_getLogs',
            ctc.rpc.construct_eth_get_logs,
            [],
            {
                'address': '0x6b175474e89094c44da98b954eedeac495271d0f',
                'start_block': start_block,
                'end_block': end_block,
            },
        ),
        (
            'eth_gasPrice',
            ctc.rpc.construct_eth_chain_id,
            [],
            {},
        ),
        (
            'eth_getBalance',
            ctc.rpc.construct_eth_get_balance,
            ['0xeb27d00030033c29307daa483171d22eb0c93342'],
            # ['0xe540c45c504b348ad4d6eb9344e6cfa07c959be6'],
            {
                'block_number': start_block,
            },
        ),
        (
            'eth_getTransactionCount',
            ctc.rpc.construct_eth_get_transaction_count,
            ['0xeb27d00030033c29307daa483171d22eb0c93342'],
            {
                'block_number': start_block,
            },
        ),
        (
            'eth_getStorageAt',
            ctc.rpc.construct_eth_get_storage_at,
            [],
            {
                'address': '0x4d9079bb4165aeb4084c526a32695dcfd2f77381',
                'position': '0x0000000000000000000000000000000000000000000000000000000000000002',  # noqa: E501
                'block_number': start_block,
            },
        ),
        (
            'eth_getCode',
            ctc.rpc.construct_eth_get_code,
            ['0x4d9079bb4165aeb4084c526a32695dcfd2f77381'],
            {},
        ),
        (
            'eth_getBlockByNumber',
            ctc.rpc.construct_eth_get_block_by_number,
            [start_block],
            {},
        ),
        (
            'eth_getBlockByHash',
            ctc.rpc.construct_eth_get_block_by_hash,
            [
                '0x3dc4ef568ae2635db1419c5fec55c4a9322c05302ae527cd40bff380c1d465dd'  # noqa: E501
            ],
            {},
        ),
        (
            'eth_getTransactionByHash',
            ctc.rpc.construct_eth_get_transaction_by_hash,
            [
                '0xd01212e8ab48d2fd2ea9c4f33f8670fd1cf0cfb09d2e3c6ceddfaf54152386e5'  # noqa: E501
            ],
            {},
        ),
        (
            'eth_getTransactionReceipt',
            ctc.rpc.construct_eth_get_transaction_receipt,
            [
                '0xd01212e8ab48d2fd2ea9c4f33f8670fd1cf0cfb09d2e3c6ceddfaf54152386e5'  # noqa: E501
            ],
            {},
        ),
        (
            'eth_call',
            ctc.rpc.construct_eth_call,
            [],
            {
                'to_address': '0x6b175474e89094c44da98b954eedeac495271d0f',
                'function_abi': {
                    'constant': True,
                    'inputs': [
                        {
                            'internalType': 'address',
                            'name': '',
                            'type': 'address',
                        }
                    ],
                    'name': 'balanceOf',
                    'outputs': [
                        {
                            'internalType': 'uint256',
                            'name': '',
                            'type': 'uint256',
                        }
                    ],
                    'payable': False,
                    'stateMutability': 'view',
                    'type': 'function',
                },
                'function_parameters': [
                    '0x5d3a536e4d6dbd6114cc1ead35777bab948e3643'
                ],
                'block_number': start_block,
            },
        ),
    ]


def get_trace_equality_tests(
    start_block: int = 10_000_000,
    end_block: int = 16_000_000,
    random_seed: flood.RandomSeed | None = None,
) -> typing.Sequence[flood.EqualityTest]:
    import ctc.rpc

    block_number = block_generators.generate_block_numbers(
        n=1,
        start_block=start_block,
        end_block=end_block,
        random_seed=random_seed,
    )[0]

    function_abi = {
        'constant': True,
        'inputs': [{'internalType': 'address', 'name': '', 'type': 'address'}],
        'name': 'balanceOf',
        'outputs': [{'internalType': 'uint256', 'name': '', 'type': 'uint256'}],
        'payable': False,
        'stateMutability': 'view',
        'type': 'function',
    }

    return [
        (
            'trace_block',
            ctc.rpc.construct_trace_block,
            [],
            {'block_number': block_number},
        ),
        (
            'trace_call',
            ctc.rpc.construct_trace_call,
            [],
            {
                'to_address': '0x6b175474e89094c44da98b954eedeac495271d0f',
                'trace_type': ['trace'],
                'function_abi': function_abi,
                'function_parameters': [
                    '0x5d3a536e4d6dbd6114cc1ead35777bab948e3643'
                ],
                'block_number': block_number,
            },
        ),
        (
            'trace_call_many',
            ctc.rpc.construct_trace_call_many,
            [],
            {
                'calls': [
                    {
                        'to_address': '0x6b175474e89094c44da98b954eedeac495271d0f',  # noqa: E501
                        'function_abi': function_abi,
                        'function_parameters': [
                            '0x5d3a536e4d6dbd6114cc1ead35777bab948e3643'
                        ],
                        'trace_type': ['trace'],
                    },
                ],
                'trace_type': ['trace'],
            },
        ),
        #     (
        #         'trace_filter',
        #         ctc.rpc.construct_trace_filter,
        #         [],
        #         {
        #             'to_addresses': ['0x6b175474e89094c44da98b954eedeac495271d0f'],  # noqa: E501
        #             'start_block': 10_000_000,
        #             'end_block': 10_000_001,
        #             'count': 10,
        #         }
        #     ),
        #     (
        #         'trace_get',
        #         ctc.rpc.construct_trace_get,
        #         [],
        #         {}
        #     ),
        #     (
        #         'trace_raw_transaction',
        #         ctc.rpc.construct_trace_raw_transaction,
        #         ['0xd01212e8ab48d2fd2ea9c4f33f8670fd1cf0cfb09d2e3c6ceddfaf54152386e5'],  # noqa: E501
        #         {'trace_type': ['trace']},
        #     ),
        (
            'trace_replay_block_transactions',
            ctc.rpc.construct_trace_replay_block_transactions,
            [],
            {
                'block_number': block_number,
                'trace_type': ['trace'],
            },
        ),
        (
            'trace_replay_transaction',
            ctc.rpc.construct_trace_replay_transaction,
            [
                '0xd01212e8ab48d2fd2ea9c4f33f8670fd1cf0cfb09d2e3c6ceddfaf54152386e5'  # noqa: E501
            ],
            {'trace_type': ['trace']},
        ),
        (
            'trace_transaction',
            ctc.rpc.construct_trace_transaction,
            [
                '0xd01212e8ab48d2fd2ea9c4f33f8670fd1cf0cfb09d2e3c6ceddfaf54152386e5'  # noqa: E501
            ],
            {},
        ),
    ]



"""wrapper around the vegeta load testing tool"""
from __future__ import annotations

import typing

from ... import spec

if typing.TYPE_CHECKING:
    import polars as pl


def run_vegeta_attack(
    *,
    url: str,
    rate: int,
    calls: typing.Sequence[typing.Any],
    duration: int,
    vegeta_kwargs: typing.Mapping[str, str | None] | None = None,
    verbose: bool = False,
    include_raw_output: bool = False,
) -> spec.LoadTestOutputDatum:
    attack = _construct_vegeta_attack(
        calls=calls,
        url=url,
        verbose=verbose,
    )
    attack_output = _vegeta_attack(
        schedule_dir=attack['schedule_dir'],
        duration=duration,
        rate=rate,
        vegeta_kwargs=vegeta_kwargs,
        verbose=verbose,
    )
    report = _create_vegeta_report(
        attack_output=attack_output,
        target_rate=rate,
        target_duration=duration,
        include_raw_output=include_raw_output,
    )
    return report


def _construct_vegeta_attack(
    calls: typing.Sequence[typing.Any],
    url: str,
    schedule_dir: str | None = None,
    verbose: bool = False,
) -> typing.Mapping[str, str]:
    import json
    import os
    import tempfile

    headers = {'Content-Type': 'application/json'}

    # determine output directory
    if schedule_dir is None:
        schedule_dir = tempfile.mkdtemp()

    # create calls file
    call_paths = []
    for c, call in enumerate(calls):
        vegeta_calls_path = os.path.join(
            schedule_dir, 'vegeta_calls_' + str(c) + '.json'
        )
        with open(vegeta_calls_path, 'w') as f:
            f.write(json.dumps(call))
        call_paths.append(vegeta_calls_path)

    # create targets specification
    vegeta_targets_path = os.path.join(schedule_dir, 'vegeta_targets')
    with open(vegeta_targets_path, 'w') as f:
        for c, call in enumerate(calls):
            f.write('POST ' + url + '\n')
            for key, value in headers.items():
                f.write(key + ': ' + value + '\n')
            f.write('@' + call_paths[c] + '\n')
            f.write('\n')

    # output summary
    if verbose:
        print('running vegeta attack...')
        print('- targets:', vegeta_targets_path)
        print('- calls:', vegeta_calls_path)

    return {
        'schedule_dir': schedule_dir,
        'vegeta_calls_path': vegeta_calls_path,
        'vegeta_targets_path': vegeta_targets_path,
    }


def _vegeta_attack(
    schedule_dir: str,
    *,
    duration: int | None = None,
    rate: int | None = None,
    max_connections: int | None = None,
    max_workers: int | None = None,
    n_cpus: int | None = None,
    report_path: str | None = None,
    vegeta_kwargs: typing.Mapping[str, str | None] | None = None,
    verbose: bool = False,
) -> bytes:
    import os
    import subprocess

    # construct command
    cmd = 'vegeta attack'
    cmd += ' -targets=' + os.path.join(schedule_dir, 'vegeta_targets')
    if rate is not None:
        cmd += ' -rate=' + str(rate)
    if duration is not None:
        cmd += ' -duration=' + str(duration) + 's'
    if max_connections is not None:
        cmd += ' -max-connections=' + str(max_connections)
    if max_workers is not None:
        cmd += ' -max-workers=' + str(max_workers)
    if vegeta_kwargs is not None:
        for key, value in vegeta_kwargs.items():
            cmd += ' -' + key
            if value is not None:
                cmd += '=' + str(value)

    if verbose:
        print('- command:', cmd)

    # run command
    return subprocess.check_output(cmd.split(' '))


def _create_vegeta_report(
    attack_output: bytes,
    target_rate: int,
    target_duration: int,
    include_raw_output: bool,
) -> spec.LoadTestOutputDatum:
    import json
    import subprocess

    cmd = 'vegeta report -type json'
    report_output = (
        subprocess.check_output(cmd.split(' '), input=attack_output)
        .decode()
        .strip()
    )
    report: spec.RawLoadTestOutputDatum = json.loads(report_output)

    if 'min' in report['latencies']:
        latency_min = report['latencies']['min'] / 1e9
    else:
        latency_min = None

    if include_raw_output:
        raw_output = encode_raw_vegeta_output(attack_output)
    else:
        raw_output = None

    return {
        'target_rate': target_rate,
        'actual_rate': report['rate'],
        'target_duration': target_duration,
        'actual_duration': report['duration'] / 1e9,
        'requests': report['requests'],
        'throughput': report['throughput'],
        'success': float(report['success']),
        'min': latency_min,
        'mean': report['latencies']['mean'] / 1e9,
        'p50': report['latencies']['50th'] / 1e9,
        'p90': report['latencies']['90th'] / 1e9,
        'p95': report['latencies']['95th'] / 1e9,
        'p99': report['latencies']['99th'] / 1e9,
        'max': report['latencies']['max'] / 1e9,
        #
        'status_codes': report['status_codes'],
        'errors': report['errors'],
        'first_request_timestamp': report['earliest'],
        'last_request_timestamp': report['latest'],
        'last_response_timestamp': report['end'],
        'final_wait_time': report['wait'] / 1e9,
        'raw_output': raw_output,
    }


#
# # output processing
#


def encode_raw_vegeta_output(raw_output: bytes) -> str:
    import base64
    import io
    import gzip

    # gzip compress
    buf = io.BytesIO()
    f = gzip.GzipFile(fileobj=buf, mode='wb')
    f.write(raw_output)
    f.close()
    compressed = buf.getvalue()

    # encode as base64
    as_base64 = base64.b64encode(compressed).decode('utf-8')

    return as_base64


def decode_raw_vegeta_output(encoded_output: str) -> bytes:
    import base64
    import io
    import gzip

    # decode from base64
    as_bytes = base64.b64decode(encoded_output.encode())

    # gzip decompress
    buf = io.BytesIO(as_bytes)
    f = gzip.GzipFile(fileobj=buf, mode='rb')
    decompressed = f.read()
    f.close()

    return decompressed


def convert_raw_vegeta_output_to_dataframe(raw_output: bytes) -> pl.DataFrame:
    import io
    import subprocess
    import polars as pl

    cmd = 'vegeta encode --to csv'
    report_output = (
        subprocess.check_output(cmd.split(' '), input=raw_output)
        .decode()
        .strip()
    )

    buf = io.StringIO()
    buf.write(report_output)
    buf.seek(0)

    schema = [
        'timestamp',
        'status_code',
        'latency',
        'bytes_out',
        'bytes_in',
        'error',
        'response',
        'name',
        'index',
        'method',
        'url',
        'response_headers',
    ]
    dtypes = {
        'bytes_in': pl.Int64,
        'bytes_out': pl.Int64,
        'error': pl.Utf8,
        'index': pl.Int64,
        'latency': pl.Int64,
        'method': pl.Utf8,
        'name': pl.Utf8,
        'response': pl.Utf8,
        'response_headers': pl.Utf8,
        'status_code': pl.Int64,
        'timestamp': pl.Int64,
        'url': pl.Utf8,
    }
    return pl.read_csv(buf, new_columns=schema, has_header=False, dtypes=dtypes)


def compute_raw_output_metrics(
    raw_output: typing.Mapping[str, pl.DataFrame],
    results: typing.Mapping[str, spec.LoadTestOutput],
) -> typing.Mapping[str, spec.LoadTestOutput]:
    import polars as pl

    metrics: typing.MutableMapping[str, spec.LoadTestOutput] = {}
    for node_name, node_results in results.items():
        node_metrics = []
        for i, (target_rate, target_duration) in enumerate(
            zip(node_results['target_rate'], node_results['target_duration'])
        ):
            sample_metrics = compute_raw_output_sample_metrics(
                df=raw_output[node_name].filter(pl.col('sample_index') == i),
                target_rate=target_rate,
                target_duration=target_duration,
            )
            sample_metrics['raw_output'] = None
            node_metrics.append(sample_metrics)
        metrics[node_name] = {
            metric: [sample_metrics[metric] for sample_metrics in node_metrics]  # type: ignore # noqa: E501
            for metric in node_results.keys()
        }
    return metrics


def compute_raw_output_sample_metrics(
    df: pl.DataFrame, target_rate: int, target_duration: int
) -> spec.LoadTestOutputDatum:
    import polars as pl
    actual_rate = (
        pl.count()
        / (pl.col('timestamp').max() - pl.col('timestamp').min())
        * 1e9
    )
    actual_rate = actual_rate.alias('actual_rate')
    successful = df.filter(pl.col('status_code') == 200)
    total_duration = (  # type: ignore
        df.select(pl.col('timestamp') + pl.col('latency')).max()
    ) - df['timestamp'].min()
    total_duration = total_duration.rows()[0][0]
    status_codes = {}
    for k, v in df['status_code'].value_counts().rows():
        status_codes[str(k)] = v
    errors = df.filter(~pl.col('error').is_null())['error'].unique().to_list()

    metrics_df = df.select(
        pl.lit(target_rate).alias('target_rate'),
        actual_rate,
        pl.lit(target_duration).alias('target_duration'),
        (pl.max('timestamp') - pl.min('timestamp')).alias('actual_duration')
        / 1e9,
        pl.count().alias('requests'),
        pl.lit(len(successful) / total_duration).alias('throughput') * 1e9,
        (len(successful) / pl.count()).alias('success'),
        pl.min('latency').alias('min') / 1e9,
        pl.mean('latency').alias('mean') / 1e9,
        pl.median('latency').alias('p50') / 1e9,
        pl.quantile('latency', 0.90).alias('p90') / 1e9,
        pl.quantile('latency', 0.95).alias('p95') / 1e9,
        pl.quantile('latency', 0.99).alias('p99') / 1e9,
        pl.max('latency').alias('max') / 1e9,
        pl.struct(**status_codes, eager=True).alias('status_codes'),
        pl.Series([errors]).alias('errors'),
        (pl.col('timestamp').min() / 1e3)
        .cast(pl.Datetime)
        .cast(str)
        .alias('first_request_timestamp'),
        (pl.col('timestamp').max() / 1e3)
        .cast(pl.Datetime)
        .cast(str)
        .alias('last_request_timestamp'),
        ((pl.col('timestamp') + pl.col('latency')) / 1e3)
        .max()
        .cast(pl.Datetime)
        .cast(str)
        .alias('last_response_timestamp'),
        (
            (pl.col('timestamp') + pl.col('latency')).max()
            - pl.max('timestamp')
        ).alias('final_wait_time')
        / 1e9,
    )

    return metrics_df.to_dicts()[0]  # type: ignore




from __future__ import annotations

import typing

import toolcli

import flood


def get_command_spec() -> toolcli.CommandSpec:
    return {
        'f': print_command,
        'help': 'print output of previous test results',
        'args': [
            {'name': 'output_dir', 'help': 'output directory of previous test'},
            {
                'name': ('-m', '--metrics'),
                'help': 'space-separated list of performance metrics',
                'nargs': '+',
            },
        ],
    }


def print_command(output_dir: str, metrics: typing.Sequence[str]) -> None:

    test_payload = flood.load_single_run_test_payload(output_dir)
    test = test_payload['test']
    results_payload = flood.load_single_run_results_payload(output_dir)
    results = results_payload['results']

    # print test summary
    flood.runners.single_runner.single_runner_summary._print_single_run_preamble_copy(
        test_name=test_payload['name'],
        rates=[subtest['rate'] for subtest in test],
        durations=[subtest['duration'] for subtest in test],
        vegeta_kwargs=[subtest['vegeta_kwargs'] for subtest in test],
        output_dir=output_dir,
    )

    # print node data
    print()
    flood.print_header('Nodes tested')
    flood.print_nodes_table(results_payload['nodes'])

    # decide metrics
    if metrics is None:
        metrics = ['success', 'throughput', 'p90']

    # print performance metrics
    print()
    print()
    flood.print_header('Summarizing performance metrics...')
    flood.print_metric_tables(
        results=results,
        metrics=metrics,
        indent=4,
    )




from __future__ import annotations

import typing

import flood
from . import equality_test_sets


def run_equality_test(
    test_name: str,
    nodes: flood.NodesShorthand,
    *,
    verbose: bool | int = True,
    random_seed: flood.RandomSeed | None = None,
    output_dir: str | None = None,
) -> None:
    import json
    import os
    import requests
    import toolstr

    nodes = flood.parse_nodes(nodes, request_metadata=True)
    for node in nodes.values():
        if node['remote'] is not None:
            raise Exception('remote not supported for equality test')
    if len(nodes) != 2:
        raise Exception('should use two nodes in equality test')

    equality_tests = equality_test_sets.get_all_equality_tests()

    # get tests
    if test_name != 'all':
        equality_tests = [t for t in equality_tests if t[0] == test_name]
        if not equality_tests:
            raise NotImplementedError(
                'no matching test found for name "' + test_name + '"'
            )

    if output_dir is None:
        import tempfile

        output_dir = tempfile.mkdtemp()

    # print preamble
    flood.print_text_box('Equality test: ' + test_name)
    flood.print_bullet(key='methods', value='')
    for test in equality_tests:
        flood.print_bullet(key=test[0], value='', colon_str='', indent=4)
    flood.print_bullet(key='output_dir', value=output_dir)
    flood.print_bullet(key='nodes', value='')
    for n, node in enumerate(nodes.values()):
        toolstr.print(
            toolstr.add_style(str(n + 1), flood.styles['metavar'])
            + '. '
            + str(node),
            indent=4,
            style=flood.styles['description'],
        )

    # run test
    successful = []
    calls = {}
    call_node_responses: typing.Any = {}
    headers = {'Content-Type': 'application/json', 'User-Agent': 'flood'}
    for test in equality_tests:
        # create call
        test_name, constructor, args, kwargs = test
        call = constructor(*args, **kwargs)
        calls[test_name] = call
        call_node_responses.setdefault(test_name, {})

        # dispatch call
        results = []
        responses = []
        for node in nodes.values():
            response = None
            try:
                response = requests.post(
                    url=node['url'], data=json.dumps(call), headers=headers
                )
                response_data = response.json()
                if 'result' in response_data:
                    result = response_data['result']
                else:
                    result = None
            except Exception:
                result = None
            responses.append(response)
            results.append(result)
            call_node_responses[test_name][node['name']] = result

        # print summary
        success = _summarize_result(
            responses=responses,
            results=results,
            nodes=nodes,
            test=test,
            call=call,
        )
        if success:
            successful.append(test_name)
    failed = [test[0] for test in equality_tests if test[0] not in successful]

    # save output file
    summary = {
        'calls': calls,
        'successful': successful,
        'failed': failed,
        'responses': call_node_responses,
    }
    file_path = os.path.join(output_dir, 'equality_results.json')
    with open(file_path, 'w') as f:
        json.dump(summary, f)

    # summarize test
    print()
    flood.print_text_box('Equality Test Summary')
    print()
    flood.print_header(
        'No differences detected (n = ' + str(len(successful)) + ')'
    )
    if len(successful) == 0:
        print('[none]')
    else:
        for name in sorted(successful):
            flood.print_bullet(key=name, value='', colon_str='')
    print()
    flood.print_header('Differences detected (n = ' + str(len(failed)) + ')')
    if len(failed) == 0:
        print('[none]')
    else:
        for name in sorted(failed):
            flood.print_bullet(key=name, value='', colon_str='')
    print()
    toolstr.print(
        'summary saved to: ' + file_path, style=flood.styles['comment']
    )


def _summarize_result(
    responses: typing.Sequence[typing.Any],
    results: typing.Sequence[typing.Any],
    nodes: flood.Nodes,
    test: flood.EqualityTest,
    call: typing.Mapping[str, typing.Any],
) -> bool:
    import toolstr

    test_name, constructor, args, kwargs = test

    if results[0] is None or not toolstr.nested_equal(results[0], results[1]):
        print()
        flood.print_text_box('Discrepancies in ' + test_name)
        print()
        flood.print_header('args')
        if len(args) > 0:
            for arg in args:
                flood.print_bullet(key=arg, value='', colon_str='')
        if len(kwargs) > 0:
            for key, value in kwargs.items():
                flood.print_bullet(key=key, value=value)
        print()
        flood.print_header('call')
        print(call)

        if any(result is None for result in results):
            print()
        for node, result, response in zip(nodes.values(), results, responses):
            if result is None:
                toolstr.print(
                    node['name'] + ' failed', style=flood.styles['title']
                )
                if response is None:
                    print('response: no response')
                elif response.status_code == 200:
                    print('response:', response.json())
                else:
                    print('response: status code ', response.status_code)

        if results[0] is None or results[1] is None:
            return False

        if len(results) == 2:
            node0, node1 = list(nodes.values())
            print()
            flood.print_header('differences in response')
            toolstr.print_nested_diff(
                lhs=results[0],
                rhs=results[1],
                lhs_name=node0['name'],
                rhs_name=node1['name'],
                styles=flood.styles,
                indent=4,
            )
        else:
            print('omitting details when >2 nodes tested')

        return False

    else:
        return True




from __future__ import annotations


raw_data_version = 'v1_0_0'

raw_data_url_template = (
    'https://datasets.paradigm.xyz/evm_samples/{network}_samples__{version}/'
)
raw_data_file_template = (
    '{network}_{datatype}_samples__{size}__{version}.parquet'
)

default_sizes = {
    'XS': 1_000,
    'S': 10_000,
    'M': 100_000,
    'L': 1_000_000,
    'XL': 10_000_000,
}

default_datatypes = [
    'contracts',
    'eoas',
    'transactions',
    'slots',
]



from __future__ import annotations

import typing

import flood


def generate_test_eth_get_transaction_by_hash(
    *,
    rates: typing.Sequence[int],
    duration: int | None = None,
    network: str,
    durations: typing.Sequence[int] | None = None,
    vegeta_kwargs: typing.Mapping[str, str | None] | None = None,
    random_seed: flood.RandomSeed | None = None,
) -> flood.LoadTest:
    n_calls = flood.estimate_call_count(
        rates=rates, duration=duration, durations=durations
    )
    calls = flood.generate_calls_eth_get_transaction_by_hash(
        n_calls=n_calls,
        network=network,
        random_seed=random_seed,
    )
    return flood.create_load_test(
        calls=calls,
        rates=rates,
        duration=duration,
        durations=durations,
    )


def generate_test_eth_get_transaction_receipt(
    *,
    rates: typing.Sequence[int],
    duration: int | None = None,
    network: str,
    durations: typing.Sequence[int] | None = None,
    vegeta_kwargs: typing.Mapping[str, str | None] | None = None,
    random_seed: flood.RandomSeed | None = None,
) -> flood.LoadTest:
    n_calls = flood.estimate_call_count(
        rates=rates, duration=duration, durations=durations
    )
    calls = flood.generate_calls_eth_get_transaction_receipt(
        n_calls=n_calls,
        network=network,
        random_seed=random_seed,
    )
    return flood.create_load_test(
        calls=calls,
        rates=rates,
        duration=duration,
        durations=durations,
    )



from __future__ import annotations

import toolcli

import flood


def get_command_spec() -> toolcli.CommandSpec:
    return {
        'f': ls_command,
        'help': 'list available tests',
    }


def ls_command() -> None:
    import toolstr

    toolstr.print_text_box(
        'Available tests',
        text_style=flood.styles['metavar'],
        style=flood.styles['content'],
    )

    print()
    toolstr.print_header(
        'Single Load Tests',
        style=flood.styles['content'],
        text_style=flood.styles['metavar'],
    )
    for test in flood.get_single_test_generators():
        toolstr.print_bullet(key=test, value='', colon_str='')
    print()
    toolstr.print_header(
        'Multi Load Tests',
        style=flood.styles['content'],
        text_style=flood.styles['metavar'],
    )
    toolstr.print_bullet(key='all', value='', colon_str='')



"""tool for benchmarking RPC endpoints"""

from .generators import *
from .runners import *
from .spec import *
from .tests import *
from .user_io import *


__version__ = '0.2.4'


def _clean_package_imports() -> None:
    """remove deep nested modules from flood namespace"""

    import sys

    flood = sys.modules['flood']
    moduletype = type(flood)
    delattr(flood, 'annotations')
    for key, value in tuple(vars(flood).items()):
        if isinstance(value, moduletype):
            name = value.__name__
            if not name.startswith('flood') or name.count('.') > 1:
                delattr(flood, key)


_clean_package_imports()



"""

TODO
--methods parameter to specify methods
"""
from __future__ import annotations

import typing

import toolcli

import flood

help_message = """Load test JSON RPC endpoints

[bold][title]Test Specification[/bold][/title]
- [metavar]TEST[/metavar] can be a template, use [metavar]flood ls[/metavar] to list test templates
- Alternatively, [metavar]TEST[/metavar] can be a directory path of a previous test
    - This will rerun this the previous test, possibly on new nodes

[bold][title]Node Specification[/bold][/title]
- Nodes are specified as a space-separated list
- Basic node syntax is [metavar]url[/metavar] or [metavar]name=url[/metavar]
- The [metavar]name[/metavar] of each node is used for benchmark summary report

[bold][title]Remote Usage[/bold][/title]
- [metavar]flood[/metavar] can be invoked on remote machines
- Use node syntax [metavar]user@remote:node_url[/metavar] or [metavar]name=user@remote:node_url[/metavar]
- Can omit the [metavar]user@[/metavar] prefix if ssh config has username specified
- [metavar]flood[/metavar] must already be installed on each remote machine

[bold][title]Parameter Randomization[/bold][/title]
- [metavar]flood[/metavar] can call each RPC method multiple times using [metavar]-n <N>[/metavar]
- For each call, parameters are randomized to minimize caching effects
- Specify random seed [metavar]-s <seed>[/metavar] for repeatable set of randomized calls"""  # noqa: E501


def get_command_spec() -> toolcli.CommandSpec:
    return {
        'f': root_command,
        'help': help_message,
        'args': [
            {
                'name': 'test',
                'help': 'test to run (use [metavar]flood ls[/metavar] for list)',  # noqa: E501
            },
            {
                'name': 'nodes',
                'nargs': '*',
                'help': 'nodes to test, see syntax above',
            },
            {
                'name': ['-s', '--seed'],
                'dest': 'random_seed',
                'type': int,
                'help': 'random seed to use, default is current timestamp',
            },
            {
                'name': ['-q', '--quiet'],
                'help': 'do not print output to [metavar]STDOUT[/metavar]',
                'action': 'store_true',
            },
            {
                'name': ['-e', '--equality'],
                'help': 'run equality test instead of load test',
                'action': 'store_true',
            },
            {
                'name': ['-m', '--mode'],
                'choices': ['stress', 'spike', 'soak'],
                'help': 'load test type: stress, spike, or soak',
            },
            {
                'name': ['-r', '--rates'],
                'nargs': '+',
                'help': 'rates to use in load test (requests per second)',
            },
            {
                'name': ['-d', '--duration'],
                'type': int,
                'help': 'amount of time to test each rate',
            },
            {
                'name': ['-o', '--output'],
                'dest': 'output_dir',
                'help': 'directory to save results, default is tmp dir',
            },
            {
                'name': ['--dry'],
                'help': 'only construct tests, do not run them',
                'action': 'store_true',
            },
            {
                'name': ['--metrics'],
                'nargs': '+',
                'help': 'space-separated list of performance metrics to show',
            },
            {
                'name': ['--no-figures'],
                'help': 'skip generating summary figures in output_dir',
                'dest': 'figures',
                'action': 'store_false',
            },
            {
                'name': ['--save-raw-output'],
                'help': 'save raw output from Vegeta',
                'action': 'store_true',
            },
            {
                'name': ['--deep-check'],
                'help': 'validate the contents of every RPC response',
                'action': 'store_true',
            },
        ],
        'examples': [
            'eth_getBlockByNumber localhost:8545',
            'eth_getLogs localhost:8545 localhost:8546 localhost:8547',
            'all client1=0.0.0.0:8545 client2=0.0.0.0:8546 --equality',
        ],
    }


def root_command(
    test: str,
    nodes: typing.Sequence[str] | None,
    output_dir: str | None,
    metrics: typing.Sequence[str],
    mode: flood.LoadTestMode | None,
    rates: typing.Sequence[int] | typing.Sequence[str] | None,
    duration: int | None,
    random_seed: int | None,
    dry: bool,
    quiet: bool,
    figures: bool,
    equality: bool,
    save_raw_output: bool,
    deep_check: bool,
) -> None:
    verbose = not quiet
    if nodes is not None and len(nodes) == 0:
        nodes = None

    if equality:
        if output_dir is not None:
            raise Exception('output_dir not used in equality test')
        if metrics is not None:
            raise Exception('metrics not used in equality test')
        if mode is not None:
            raise Exception('metrics not used in equality test')
        if rates is not None:
            raise Exception('rates not used in equality test')
        if duration is not None:
            raise Exception('duration not used in equality test')
        if dry:
            raise Exception('dry not used in equality test')
        if not figures:
            raise Exception('figures not used in equality test')
        flood.run_equality_test(
            test_name=test,
            nodes=nodes,
            random_seed=random_seed,
            verbose=verbose,
        )

    else:
        if rates is not None:
            rates = [int(rate) for rate in rates]
        flood.run(
            test_name=test,
            mode=mode,
            nodes=nodes,
            metrics=metrics,
            random_seed=random_seed,
            verbose=verbose,
            rates=rates,
            duration=duration,
            dry=dry,
            output_dir=output_dir or True,
            figures=figures,
            include_raw_output=save_raw_output,
            deep_check=deep_check,
        )




from __future__ import annotations

import typing
import flood


def generate_slots(
    n: int,
    network: str,
    random_seed: flood.RandomSeed | None = None,
) -> typing.Sequence[tuple[str, str]]:
    return flood.load_samples(
        network=network,
        datatype='slots',
        n=n,
        random_seed=random_seed,
    )



from __future__ import annotations

import typing

from flood import spec
from .. import rng_utils


def generate_block_numbers(
    n: int,
    start_block: int,
    end_block: int,
    *,
    replace: bool = False,
    sort: bool = False,
    random_seed: spec.RandomSeed | None = None,
    network: str | None = None,
) -> typing.Sequence[int]:
    import numpy as np

    # seed a generator
    rng = rng_utils.get_rng(random_seed=random_seed)

    # generate blocks
    all_blocks = np.arange(start_block, end_block + 1)
    chosen_array = rng.choice(all_blocks, size=n, replace=replace)
    chosen: list[int] = chosen_array.tolist()

    # sort
    if sort:
        chosen = sorted(chosen)

    return chosen


def generate_block_hashes(
    n: int,
    network: str | None = None,
    random_seed: spec.RandomSeed | None = None,
) -> typing.Sequence[str]:
    raise NotImplementedError()


def generate_block_ranges(
    *,
    n: int,
    range_size: int,
    start_block: int,
    end_block: int,
    non_overlapping: bool = True,
    sort: bool = False,
    n_attempts: int = 1_000_000,
    random_seed: spec.RandomSeed | None = None,
    method: str = 'strides',
    network: str | None = None,
) -> typing.Sequence[tuple[int, int]]:
    if method == 'strides':
        return _generate_block_ranges_strides(
            n=n,
            range_size=range_size,
            start_block=start_block,
            end_block=end_block,
            non_overlapping=non_overlapping,
            sort=sort,
            n_attempts=n_attempts,
            random_seed=random_seed,
        )
    elif method == 'individual':
        return _generate_block_ranges_individual(
            n=n,
            range_size=range_size,
            start_block=start_block,
            end_block=end_block,
            non_overlapping=non_overlapping,
            sort=sort,
            n_attempts=n_attempts,
            random_seed=random_seed,
        )
    else:
        raise Exception('unknown method: ' + str(method))


def _generate_block_ranges_strides(
    *,
    n: int,
    range_size: int,
    start_block: int,
    end_block: int,
    non_overlapping: bool = True,
    sort: bool = False,
    n_attempts: int = 1_000_000,
    random_seed: spec.RandomSeed | None = None,
) -> typing.Sequence[tuple[int, int]]:
    import numpy as np

    # seed a generator
    rng = rng_utils.get_rng(random_seed=random_seed)

    block_ranges: list[tuple[int, int]] = []
    while len(block_ranges) < n:
        random_phases = list(range(range_size))
        rng.shuffle(random_phases)
        for random_phase in random_phases:
            strides = np.arange(
                start_block + random_phase, end_block, range_size
            )
            rng.shuffle(strides)
            n_missing = n - len(block_ranges)
            if n_missing > 0:
                for stride in strides[:n_missing]:
                    block_ranges.append((stride, stride + range_size))
            else:
                break

    return block_ranges


def _generate_block_ranges_individual(
    *,
    n: int,
    range_size: int,
    start_block: int,
    end_block: int,
    non_overlapping: bool = True,
    sort: bool = False,
    n_attempts: int = 1_000_000,
    random_seed: spec.RandomSeed | None = None,
) -> typing.Sequence[tuple[int, int]]:
    # seed a generator
    rng = rng_utils.get_rng(random_seed=random_seed)

    # create starting sample
    start_blocks = generate_block_numbers(
        n=n,
        start_block=start_block,
        end_block=end_block,
        replace=non_overlapping,
    )
    candidates = iter(start_blocks)

    # create ranges
    attempt = 0
    ranges: set[tuple[int, int]] = set()
    while len(ranges) < n:
        # record attempt
        attempt += 1
        if attempt > n_attempts:
            raise Exception(
                'could not find block range set after '
                + str(n_attempts)
                + ' attempts'
            )

        # create new candidate
        try:
            start = next(candidates)
        except StopIteration:
            start = rng.integers(start_block, end_block)
        end = start + range_size
        candidate = (start, end)

        # check whether candidate is valid
        if candidate in ranges:
            continue
        if end > end_block:
            continue
        if non_overlapping:
            skip = False
            for other_start, other_end in ranges:
                if other_start <= end and other_end >= start:
                    skip = True
                    break
            if skip:
                continue

        ranges.add(candidate)

    # sort
    if sort:
        return sorted(ranges)
    else:
        ranges_list = list(ranges)
        rng.shuffle(ranges_list)
        return ranges_list



from __future__ import annotations

import typing

import flood


def estimate_call_count(
    *,
    rates: typing.Sequence[int],
    duration: int | None = None,
    durations: typing.Sequence[int] | None = None,
    n_repeats: int | None = None,
) -> int:
    if duration is not None:
        n_calls = sum(rate * duration for rate in rates)
    elif durations is not None:
        if len(rates) != len(durations):
            raise Exception('different number of rates vs durations')
        n_calls = sum(
            rate * duration for rate, duration in zip(rates, durations)
        )
    else:
        raise Exception('should specify duration or durations')

    if n_repeats is not None:
        n_calls *= n_repeats

    return n_calls


def create_load_test(
    calls: typing.Sequence[typing.Any],
    rates: typing.Sequence[int],
    duration: int | None = None,
    durations: typing.Sequence[int] | None = None,
    vegeta_kwargs: flood.VegetaKwargs
    | typing.Sequence[flood.VegetaKwargs]
    | None = None,
    repeat_calls: bool = False,
) -> flood.LoadTest:
    # validate inputs
    if len(rates) == 0:
        raise Exception('must specify at least one rate')

    # pluralize singular durations
    if durations is None:
        if duration is None:
            raise Exception('must specify duration or durations')
        durations = [duration] * len(rates)
    assert len(durations) == len(rates)

    # pluralize singular vegeta kwargs
    if vegeta_kwargs is None:
        vegeta_kwargs = {}
    if isinstance(vegeta_kwargs, dict):
        vegeta_kwargs = [vegeta_kwargs] * len(rates)
    if not isinstance(vegeta_kwargs, list):
        raise Exception('invalid input')

    # partition calls into individual attacks
    if not repeat_calls:
        attacks_calls: typing.MutableSequence[typing.Sequence[flood.Call]] = []
        calls_iter = iter(calls)
        for rate, duration in zip(rates, durations):
            n_attack_calls = rate * duration
            attack_calls = []
            for i in range(n_attack_calls):
                attack_calls.append(next(calls_iter))
            attacks_calls.append(attack_calls)
    else:
        attacks_calls = [calls] * len(rates)
    assert len(attacks_calls) == len(rates)

    # create load tests
    load_test: list[flood.VegetaAttack] = []
    for rate, duration, a_calls, attack_kwargs in zip(
        rates, durations, attacks_calls, vegeta_kwargs
    ):
        attack: flood.VegetaAttack = {
            'rate': rate,
            'duration': duration,
            'calls': a_calls,
            'vegeta_kwargs': attack_kwargs,
        }
        load_test.append(attack)

    return load_test



from __future__ import annotations

import pytest

import flood


names = [None, 'test_node', 'backup_node']

valid_hosts = ['localhost', '61.4.1.34', '12.53.23.35', 'some_provider.com']

valid_ports = ['', ':8545', ':80']

url_prefixes = ['', 'http://', 'https://']

url_suffixes = ['/provider_key', '']

remote_clients = [None]
for host in valid_hosts:
    remote_clients.append(host)
    remote_clients.append('username@' + host)


@pytest.mark.parametrize('name', names)
@pytest.mark.parametrize('remote_client', remote_clients)
@pytest.mark.parametrize('url_prefix', url_prefixes)
@pytest.mark.parametrize('node_host', valid_hosts)
@pytest.mark.parametrize('port', valid_ports)
@pytest.mark.parametrize('url_suffix', url_suffixes)
def test_parse_node(
    name,
    remote_client,
    url_prefix,
    node_host,
    port,
    url_suffix,
):
    node_url = url_prefix + node_host + port + url_suffix
    node_str = node_url
    if remote_client is not None:
        node_str = remote_client + ':' + node_str
    unnamed_node_str = node_str
    if name is not None:
        node_str = name + '=' + node_str

    parsed = flood.parse_node(node_str, request_metadata=False)

    if name is not None:
        assert parsed['name'] == name
    else:
        assert parsed['name'] == unnamed_node_str

    pieces = node_url.split(':')[0].split('.')
    is_ip = len(pieces) == 4 and all(
        piece.isdecimal() for piece in pieces
    )

    if url_prefix != '':
        assert parsed['url'] == node_url
    elif (
        node_url.startswith('localhost') or is_ip
    ):
        assert parsed['url'] == 'http://' + node_url
    else:
        assert parsed['url'] == 'https://' + node_url

    assert parsed['remote'] == remote_client




from __future__ import annotations

import typing

import flood
from ..single_runner import single_runner_io


_path_templates = {
    'multi_indicator': 'multi_test.fyi',
}


def _save_multi_run_tests(
    output_dir: str,
    tests: typing.Mapping[str, flood.LoadTest],
) -> None:
    import os

    # make output dir
    if not os.path.isdir(output_dir):
        if os.path.exists(output_dir):
            raise Exception('output must be a directory path')
        else:
            os.makedirs(output_dir, exist_ok=True)

    # create individual tests
    for test_name, test in tests.items():
        single_runner_io._save_single_run_test(
            test_name=test_name,
            output_dir=os.path.join(output_dir, test_name),
            test=test,
        )

    # create multitest indicator file
    indicator_path = os.path.join(
        output_dir, _path_templates['multi_indicator']
    )
    with open(indicator_path, 'a'):
        pass


def _load_multi_run_tests(
    output_dir: str,
) -> typing.Mapping[str, flood.LoadTest]:
    raise NotImplementedError()





import flood


def test_local_node_1(flood_test_local_node_1):
    result = flood.parse_node(flood_test_local_node_1)
    assert result['remote'] is None


def test_local_node_2(flood_test_local_node_2):
    result = flood.parse_node(flood_test_local_node_2)
    assert result['remote'] is None


def test_remote_node_1(flood_test_remote_node_1):
    result = flood.parse_node(flood_test_remote_node_1)
    assert result['remote'] is not None


def test_remote_node_2(flood_test_remote_node_2):
    result = flood.parse_node(flood_test_remote_node_2)
    assert result['remote'] is not None




from __future__ import annotations

import typing

import toolcli

import flood


def get_command_spec() -> toolcli.CommandSpec:
    return {
        'f': download_samples_command,
        'help': 'download raw EVM samples for constructing calls',
        'args': [
            {
                'name': ('-n', '--network'),
                'help': 'network, default ethereum',
            },
            {
                'name': ('-s', '--sizes'),
                'help': 'sample sizes, one of {XS, S, M, L, XL, all}, default L',  # noqa: E501
                'nargs': '+',
            },
            {
                'name': ('-o', '--output-dir'),
                'help': 'download location, default FLOOD_SAMPLES_DIR or cwd',
            },
            {
                'name': ('-d', '--datatypes'),
                'help': 'datatypes to sample, default all',
            },
            {
                'name': ('-m', '--missing'),
                'help': 'only download missing files',
                'action': 'store_true',
            },
        ],
    }


def download_samples_command(
    network: str | None,
    sizes: typing.Sequence[str] | None,
    output_dir: str | None,
    datatypes: typing.Sequence[str] | None,
    missing: bool,
) -> None:
    if network is None:
        network = 'ethereum'
    if sizes is None:
        sizes = ['L']
    if sizes == 'all':
        sizes = list(flood.default_sizes.keys())
    if output_dir is None:
        output_dir = flood.get_flood_samples_dir()

    flood.download_raw_data(
        network=network,
        sizes=sizes,
        datatypes=datatypes,
        output_dir=output_dir,
        only_missing=missing,
    )



from __future__ import annotations

import typing

import flood


def generate_test_eth_get_logs(
    *,
    rates: typing.Sequence[int],
    network: str,
    duration: int | None = None,
    durations: typing.Sequence[int] | None = None,
    vegeta_kwargs: typing.Mapping[str, str | None] | None = None,
    random_seed: flood.RandomSeed | None = None,
) -> flood.LoadTest:
    n_calls = flood.estimate_call_count(
        rates=rates, duration=duration, durations=durations
    )
    calls = flood.generate_calls_eth_get_logs(
        n_calls,
        network=network,
        random_seed=random_seed,
    )
    return flood.create_load_test(
        calls=calls,
        rates=rates,
        duration=duration,
        durations=durations,
    )


# def generate_tests_eth_get_logs_by_contract(
#     url: str,
#     rates: typing.Sequence[int],
#     duration: int,
#     range_size: int,
# ) -> typing.Mapping[str, flood.LoadTest]:
#     """test: Transfers of USDC vs DAI vs LUSD"""

#     n_calls = flood.estimate_call_count(rates=rates, duration=duration)

#     block_ranges = flood.generate_block_ranges(
#         start_block=12_178_594,
#         end_block=17_000_000,
#         n=n_calls,
#         range_size=range_size,
#     )

#     tests: typing.MutableMapping[str, flood.LoadTest] = {}
#     for name, contract_address in contracts.items():
#         calls = flood.generate_calls_eth_get_logs(
#             contract_address=contract_address,
#             topics=[event_hashes['Transfer']],
#             block_ranges=block_ranges,
#         )
#         tests[name] = {
#             'url': url,
#             'rates': rates,
#             'duration': duration,
#             'calls': calls,
#         }

#     return tests


# def generate_tests_eth_get_logs_by_block_range_size(
#     url: str,
#     rates: typing.Sequence[int],
#     duration: int,
#     range_sizes: typing.Sequence[int],
# ) -> typing.Mapping[str, flood.LoadTest]:
#     """test: tiny vs small vs medium vs large block ranges"""

#     n_calls = flood.estimate_call_count(rates=rates, duration=duration)

#     tests: typing.MutableMapping[str, flood.LoadTest] = {}
#     for range_size in range_sizes:
#         block_ranges = flood.generate_block_ranges(
#             start_block=12_178_594,
#             end_block=17_000_000,
#             n=n_calls,
#             range_size=range_size,
#         )
#         calls = flood.generate_calls_eth_get_logs(
#             contract_address=contracts['USDC'],
#             topics=[event_hashes['Transfer']],
#             block_ranges=block_ranges,
#         )
#         name = str(range_size) + '_blocks'
#         tests[name] = {
#             'url': url,
#             'rates': rates,
#             'duration': duration,
#             'calls': calls,
#         }

#     return tests


# def generate_tests_eth_get_logs_by_block_age(
#     url: str,
#     rates: typing.Sequence[int],
#     duration: int,
#     block_bounds: typing.Mapping[str, tuple[int, int]],
# ) -> typing.Mapping[str, flood.LoadTest]:
#     """test old vs new blocks"""

#     n_calls = flood.estimate_call_count(rates=rates, duration=duration)

#     tests: typing.MutableMapping[str, flood.LoadTest] = {}
#     for name, (start_block, end_block) in block_bounds.items():
#         block_ranges = flood.generate_block_ranges(
#             start_block=start_block,
#             end_block=end_block,
#             n=n_calls,
#             range_size=100,
#         )
#         calls = flood.generate_calls_eth_get_logs(
#             contract_address=contracts['USDC'],
#             topics=[event_hashes['Transfer']],
#             block_ranges=block_ranges,
#         )
#         tests[name] = {
#             'url': url,
#             'rates': rates,
#             'duration': duration,
#             'calls': calls,
#         }

#     return tests



from __future__ import annotations

import typing

import flood
from . import single_runner_io
from . import single_runner_summary


def _run_single(
    *,
    test_name: str,
    rerun_of: str | None = None,
    test: flood.LoadTest | None = None,
    nodes: flood.NodesShorthand,
    random_seed: flood.RandomSeed | None = None,
    rates: typing.Sequence[int] | None = None,
    duration: int | None = None,
    durations: typing.Sequence[int] | None = None,
    mode: flood.LoadTestMode | None = None,
    vegeta_kwargs: flood.VegetaKwargsShorthand | None = None,
    dry: bool,
    output_dir: str | None = None,
    figures: bool,
    metrics: typing.Sequence[str] | None = None,
    verbose: bool | int,
    include_raw_output: bool = False,
    deep_check: bool = False,
) -> None:
    if deep_check:
        include_raw_output = True

    # get test parameters
    rates, durations, vegeta_kwargs = _get_single_test_parameters(
        test=test,
        rates=rates,
        duration=duration,
        durations=durations,
        mode=mode,
    )

    # print preamble
    if verbose:
        single_runner_summary._print_single_run_preamble_copy(
            test_name=test_name,
            rerun_of=rerun_of,
            rates=rates,
            durations=durations,
            vegeta_kwargs=vegeta_kwargs,
            output_dir=output_dir,
        )

    # parse nodes
    nodes = flood.parse_nodes(nodes, verbose=verbose, request_metadata=True)

    # generate test and save to disk
    if test is None:
        test = flood.generate_test(
            test_name=test_name,
            rates=rates,
            durations=durations,
            vegeta_kwargs=vegeta_kwargs,
            network=flood.parse_nodes_network(nodes),
            random_seed=random_seed,
            output_dir=output_dir,
        )

    # skip dry run
    if dry:
        print()
        print('[dry run, exitting]')
        return

    # run tests
    if verbose:
        single_runner_summary._print_run_start()
    results = flood.run_load_tests(
        nodes=nodes,
        test=test,
        verbose=verbose,
        include_raw_output=include_raw_output,
    )

    # output results to file
    if output_dir is not None:
        single_runner_io._save_single_run_results(
            output_dir=output_dir,
            test=test,
            nodes=nodes,
            results=results,
            figures=figures,
            test_name=test_name,
        )

    # perform deep check
    if deep_check:
        _perform_deep_check(results, verbose=verbose)

    # print summary
    if verbose:
        single_runner_summary._print_single_run_conclusion_copy(
            output_dir=output_dir,
            results=results,
            metrics=metrics,
            verbose=verbose,
            figures=figures,
        )


#
# # helper functions
#


def _perform_deep_check(
    results: typing.Mapping[str, flood.LoadTestOutput],
    verbose: bool | int = False,
) -> None:
    """
    check that responses for each success are
    1) well-formed json
    and 2) error = None
    """
    import base64
    import json

    errors = False

    raw_output = flood.load_single_run_raw_output(results=results)
    for result_name, result in raw_output.items():
        for status_code, response in zip(
            result['status_code'], result['response'].to_list()
        ):
            if status_code == 200:
                try:
                    decoded = json.loads(base64.b64decode(response))
                    if decoded.get('result') is None:
                        errors = True
                except Exception:
                    errors = True

            if errors:
                break
        if errors:
            break

    if verbose:
        if errors:
            print()
            print('[deep check passed]')
        else:
            print('[deep check failed]')

    if errors:
        raise Exception('some calls that were reported successful had bad data')


def _get_single_test_parameters(
    test: flood.LoadTest | None = None,
    rates: typing.Sequence[int] | None = None,
    duration: int | None = None,
    durations: typing.Sequence[int] | None = None,
    mode: flood.LoadTestMode | None = None,
    vegeta_kwargs: flood.VegetaKwargsShorthand | None = None,
) -> tuple[
    typing.Sequence[int],
    typing.Sequence[int],
    flood.VegetaKwargsShorthand | None,
]:
    if test is not None:
        test_data = flood.parse_test_data(test=test)
        rates = test_data['rates']
        durations = test_data['durations']
        vegeta_kwargs = test_data['vegeta_kwargs']
    else:
        rates, durations = flood.generate_timings(
            rates=rates,
            duration=duration,
            durations=durations,
            mode=mode,
        )
    return rates, durations, vegeta_kwargs




from .address_test_generators import *
from .block_test_generators import *
from .contract_test_generators import *
from .generic_test_generators import *
from .log_test_generators import *
from .multi_test_generators import *
from .transaction_test_generators import *
from .trace_test_generators import *



from __future__ import annotations

import typing

import flood
from flood import spec


def get_ctc_alias_url(url: str) -> str | None:
    try:
        import ctc.config

        ctc_providers = ctc.config.get_providers()
        if url in ctc_providers:
            url = ctc_providers[url]['url']
        return url
    except ImportError:
        return None


def parse_nodes(
    nodes: spec.NodesShorthand,
    *,
    verbose: bool | int = False,
    request_metadata: bool = False,
) -> typing.Mapping[str, spec.Node]:
    """parse given nodes according to input specification"""
    if verbose:
        flood.print_header('Gathering node data...')

    new_nodes: typing.MutableMapping[str, spec.Node] = {}
    if isinstance(nodes, list):
        import multiprocessing

        with multiprocessing.Pool() as pool:
            results = pool.map(parse_node, nodes)
        new_nodes = {result['name']: result for result in results}
    elif isinstance(nodes, dict):
        for key, value in nodes.items():
            new_nodes[key] = value
    else:
        raise Exception('invalid format for nodes')

    if verbose:
        print_nodes_table(new_nodes)

    return new_nodes


def print_nodes_table(nodes: typing.Mapping[str, spec.Node]) -> None:
    rows = []
    for node in nodes.values():
        url = node['url']
        if node['remote'] is not None:
            url = node['remote'] + '\n' + url
        row = [
            node['name'],
            url,
            node['client_version'],
        ]
        rows.append(row)
    labels = ['node', 'url', 'metadata']
    print()
    flood.print_multiline_table(
        rows=rows,
        labels=labels,
        indent=4,
    )


def _get_node_str(node: flood.Node) -> str:
    node_str = '"' + node['name'] + '", url=' + node['url']
    remote = node['remote']
    if remote is not None:
        node_str += ' remote=' + remote
    if node['client_version'] is not None:
        node_str = node_str + ' version=' + node['client_version']
    return node_str


def parse_node(
    node: str | spec.Node, request_metadata: bool = True
) -> spec.Node:
    """parse node according to input specification"""
    prefixes = ['http', 'https', 'ws', 'wss']

    if isinstance(node, dict):
        return node
    elif isinstance(node, str):
        # parse name
        if '=' in node:
            name, url = node.split('=', 1)
        else:
            name = node
            url = node

        # check if node is in ctc aliases
        alias_url = get_ctc_alias_url(url)
        if alias_url is not None:
            url = alias_url

        # parse remote and url
        if ':' in url:
            head, tail = url.split(':', 1)
            if head in prefixes:
                remote = None
                url = head + ':' + tail
            else:
                if tail.split('/')[0].isdecimal():
                    remote = None
                    url = url
                else:
                    remote = head
                    url = tail
        else:
            remote = None

        # add missing prefix
        if not any(url.startswith(prefix) for prefix in prefixes):
            # check if is ip
            pieces = url.split(':')[0].split('.')
            is_ip = len(pieces) == 4 and all(
                piece.isdecimal() for piece in pieces
            )

            # add prefix
            if url.startswith('localhost') or is_ip:
                url = 'http://' + url
            else:
                url = 'https://' + url

        if request_metadata:
            # get client version
            client_version = get_node_client_version(url=url, remote=remote)

            # get network
            network = 'ethereum'

        else:
            client_version = None
            network = None

        return {
            'name': name,
            'url': url,
            'remote': remote,
            'client_version': client_version,
            'network': network,
        }

    else:
        raise Exception('invalid node format')


def get_node_client_version(url: str, remote: str | None = None) -> str | None:
    try:
        if remote is None:
            import ctc.rpc

            info: str = ctc.rpc.sync_web3_client_version(
                context={'provider': url}
            )
            return info
        else:
            import json
            import subprocess

            cmd = [
                """ssh""",
                remote,
                """curl -X POST -H 'Content-Type: application/json' -d '{\"jsonrpc\": \"2.0\", \"method\": \"web3_clientVersion\", \"params\": [], \"id\": 1}' """  # noqa: E501
                + url,
            ]
            output = subprocess.check_output(cmd, stderr=subprocess.DEVNULL)
            response = json.loads(output)
            info = response['result']
            return info
    except Exception:
        return None


def parse_test_data(test: spec.LoadTest) -> spec.LoadTestColumnWise:
    rates = []
    durations = []
    vegeta_kwargs = []
    calls = []
    for attack in test:
        rates.append(attack['rate'])
        durations.append(attack['duration'])
        vegeta_kwargs.append(attack['vegeta_kwargs'])
        calls.append(attack['calls'])
    return {
        'rates': rates,
        'durations': durations,
        'vegeta_kwargs': vegeta_kwargs,
        'calls': calls,
    }


def parse_nodes_network(nodes: typing.Mapping[str, spec.Node]) -> str:
    if len(nodes) == 0:
        raise Exception('no nodes, cannot get network')
    networks = [node['network'] for node in nodes.values()]
    if len(set(networks)) != 1:
        raise Exception('multiple networks')
    if networks[0] is None:
        raise Exception('network could not be determined')
    network = networks[0]
    if isinstance(network, str):
        return network
    else:
        raise Exception('need str network')



from __future__ import annotations

import typing

import toolcli

import flood


def get_command_spec() -> toolcli.CommandSpec:
    return {
        'f': samples_collect_command,
        'help': 'collect raw EVM samples using Paradigm Data Portal datasets',
        'args': [
            {
                'name': ('-n', '--network'),
                'help': 'network, default ethereum',
            },
            {
                'name': ('-s', '--sizes'),
                'help': 'sample sizes of {XS, S, M, L, XL}, default all',
            },
            {
                'name': ('-o', '--output-dir'),
                'help': 'download location, default FLOOD_SAMPLES_DIR or cwd',
            },
            {
                'name': ('-d', '--datatypes'),
                'help': 'datatypes to sample, default all',
            },
        ],
    }


def samples_collect_command(
    network: str | None,
    sizes: typing.Sequence[str] | None,
    output_dir: str | None,
    datatypes: typing.Sequence[str] | None,
) -> None:
    if output_dir is None:
        output_dir = flood.get_flood_samples_dir()
    if network is None:
        network = 'ethereum'
    flood.create_samples_dataset(
        output_dir=output_dir,
        network=network,
        sizes=sizes,
        datatypes=datatypes,
    )



from __future__ import annotations

import typing

import flood

if typing.TYPE_CHECKING:
    import polars as pl


_path_templates = {
    'single_run_test': '{output_dir}/test.json',
    'single_run_results': '{output_dir}/results.json',
    'single_run_figures_dir': '{output_dir}/figures',
}


def get_single_run_test_path(output_dir: str) -> str:
    return _path_templates['single_run_test'].format(output_dir=output_dir)


def get_single_run_results_path(output_dir: str) -> str:
    return _path_templates['single_run_results'].format(output_dir=output_dir)


def get_single_run_figures_path(output_dir: str) -> str:
    return _path_templates['single_run_figures_dir'].format(
        output_dir=output_dir
    )


def _save_single_run_test(
    test_name: str,
    output_dir: str,
    test: flood.LoadTest,
) -> None:
    import os
    import json

    if not os.path.isdir(output_dir):
        if os.path.exists(output_dir):
            raise Exception('output must be a directory path')
        else:
            os.makedirs(output_dir, exist_ok=True)

    path = _path_templates['single_run_test'].format(output_dir=output_dir)
    payload: flood.SingleRunTestPayload = {
        'version': flood.__version__,
        'type': 'single_test',
        'name': test_name,
        'test': test,
    }
    with open(path, 'w') as f:
        json.dump(payload, f)


def _save_single_run_results(
    *,
    output_dir: str,
    test: flood.LoadTest,
    nodes: flood.Nodes,
    results: typing.Mapping[str, flood.LoadTestOutput],
    figures: bool,
    test_name: str,
) -> None:
    import os
    import json

    if not os.path.isdir(output_dir):
        if os.path.exists(output_dir):
            raise Exception('output must be a directory path')
        else:
            os.makedirs(output_dir)

    path = _path_templates['single_run_results'].format(output_dir=output_dir)
    payload: flood.SingleRunResultsPayload = {
        'version': flood.__version__,
        'type': 'single_test',
        'nodes': nodes,
        'results': results,
    }
    with open(path, 'w') as f:
        json.dump(payload, f)

    if figures:
        figures_dir = get_single_run_figures_path(output_dir=output_dir)
        colors = flood.get_nodes_plot_colors(nodes=nodes)
        flood.plot_load_test_results(
            outputs=results,
            test_name=test_name,
            output_dir=figures_dir,
            colors=colors,
        )


def load_single_run_test_payload(
    path_spec: str,
) -> flood.SingleRunTestPayload:
    import os
    import json

    if os.path.isfile(path_spec):
        path = path_spec
    else:
        path = get_single_run_test_path(path_spec)
    with open(path) as f:
        test: flood.SingleRunTestPayload = json.load(f)
    return test


def load_single_run_test(
    output_dir: str,
) -> flood.LoadTest:
    payload = load_single_run_test_payload(output_dir)
    return payload['test']


def load_single_run_results_payload(
    output_dir: str,
) -> flood.SingleRunResultsPayload:
    import json

    path = get_single_run_results_path(output_dir=output_dir)
    with open(path) as f:
        results: flood.SingleRunResultsPayload = json.load(f)
    return results


def load_single_run_results(
    output_dir: str,
) -> typing.Mapping[str, flood.LoadTestOutput]:
    payload = load_single_run_results_payload(output_dir=output_dir)
    return payload['results']


def load_single_run_raw_output(
    *,
    output_dir: str | None = None,
    results: typing.Mapping[str, flood.LoadTestOutput] | None = None,
    sample_index: int | None = None,
) -> typing.Mapping[str, pl.DataFrame]:
    import polars as pl

    if results is None:
        if output_dir is None:
            raise Exception('must specify output_dir or results')
        results = load_single_run_results(output_dir=output_dir)
    node_dfs = {}
    for node_name, node_results in results.items():
        raw_output = node_results['raw_output']
        if raw_output is None:
            raise Exception('raw_outputs were not saved for test')
        else:
            if sample_index is not None:
                indices = [sample_index]
            else:
                indices = list(range(len(raw_output)))
            dfs = []
            for index in indices:
                item = raw_output[index]
                if item is None:
                    raise Exception('raw_outputs were not saved for test')
                decoded = flood.decode_raw_vegeta_output(item)
                df = flood.convert_raw_vegeta_output_to_dataframe(decoded)
                df = df.with_columns(pl.lit(index).alias('sample_index'))
                dfs.append(df)
            node_dfs[node_name] = pl.concat(dfs)
    return node_dfs

